from pathlib import Path

import pytest
from typer import Typer
from typer.testing import CliRunner

from strawberry.cli.commands.codegen import ConsolePlugin
from strawberry.codegen import CodegenFile, CodegenResult, QueryCodegenPlugin
from strawberry.codegen.types import GraphQLOperation, GraphQLType


class ConsoleTestPlugin(ConsolePlugin):
    def on_end(self, result: CodegenResult):
        result.files[0].path = "renamed.py"

        return super().on_end(result)


class QueryCodegenTestPlugin(QueryCodegenPlugin):
    def generate_code(
        self, types: list[GraphQLType], operation: GraphQLOperation
    ) -> list[CodegenFile]:
        return [
            CodegenFile(
                path="test.py",
                content=f"# This is a test file for {operation.name}",
            )
        ]


class EmptyPlugin(QueryCodegenPlugin):
    def generate_code(
        self, types: list[GraphQLType], operation: GraphQLOperation
    ) -> list[CodegenFile]:
        return [
            CodegenFile(
                path="test.py",
                content="# Empty",
            )
        ]


@pytest.fixture
def query_file_path(tmp_path: Path) -> Path:
    output_path = tmp_path / "query.graphql"
    output_path.write_text(
        """
        query GetUser {
            user {
                name
            }
        }
        """
    )
    return output_path


@pytest.fixture
def query_file_path2(tmp_path: Path) -> Path:
    output_path = tmp_path / "query2.graphql"
    output_path.write_text(
        """
        query GetUser {
            user {
                name
            }
        }
        """
    )
    return output_path


def test_codegen(
    cli_app: Typer, cli_runner: CliRunner, query_file_path: Path, tmp_path: Path
):
    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "-p",
            "tests.cli.test_codegen:QueryCodegenTestPlugin",
            "-o",
            str(tmp_path),
            "--schema",
            selector,
            str(query_file_path),
        ],
    )

    assert result.exit_code == 0

    code_path = tmp_path / "test.py"

    assert code_path.exists()
    assert code_path.read_text() == "# This is a test file for GetUser"


def test_codegen_multiple_files(
    cli_app: Typer,
    cli_runner: CliRunner,
    query_file_path: Path,
    query_file_path2: Path,
    tmp_path: Path,
):
    expected_paths = [
        tmp_path / "query.py",
        tmp_path / "query2.py",
        tmp_path / "query.ts",
        tmp_path / "query2.ts",
    ]
    for path in expected_paths:
        assert not path.exists()

    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "-p",
            "python",
            "-p",
            "typescript",
            "-o",
            str(tmp_path),
            "--schema",
            selector,
            str(query_file_path),
            str(query_file_path2),
        ],
    )

    assert result.exit_code == 0

    for path in expected_paths:
        assert path.exists()
        assert " GetUserResult" in path.read_text()


def test_codegen_pass_no_query(cli_app: Typer, cli_runner: CliRunner, tmp_path: Path):
    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "-p",
            "tests.cli.test_codegen:EmptyPlugin",
            "-o",
            str(tmp_path),
            "--schema",
            selector,
        ],
    )

    assert result.exit_code == 0


def test_codegen_passing_plugin_symbol(
    cli_app: Typer, cli_runner: CliRunner, query_file_path: Path, tmp_path: Path
):
    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "-p",
            "tests.cli.test_codegen:EmptyPlugin",
            "-o",
            str(tmp_path),
            "--schema",
            selector,
            str(query_file_path),
        ],
    )

    assert result.exit_code == 0

    code_path = tmp_path / "test.py"

    assert code_path.exists()
    assert code_path.read_text() == "# Empty"


def test_codegen_returns_error_when_symbol_does_not_exist(
    cli_app: Typer, cli_runner: CliRunner, query_file_path: Path, tmp_path: Path
):
    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "-p",
            "tests.cli.test_codegen:SomePlugin",
            "--schema",
            selector,
            "-o",
            str(tmp_path),
            str(query_file_path),
        ],
    )

    assert result.exit_code == 1
    assert result.exception
    assert result.exception.args == (
        "module 'tests.cli.test_codegen' has no attribute 'SomePlugin'",
    )


def test_codegen_returns_error_when_module_does_not_exist(
    cli_app: Typer, cli_runner: CliRunner, query_file_path: Path, tmp_path: Path
):
    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "-p",
            "fake_module_plugin",
            "--schema",
            selector,
            "-o",
            str(tmp_path),
            str(query_file_path),
        ],
    )

    assert result.exit_code == 1
    assert "Error: Plugin fake_module_plugin not found" in result.output


def test_codegen_returns_error_when_does_not_find_plugin(
    cli_app: Typer, cli_runner: CliRunner, query_file_path: Path, tmp_path: Path
):
    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "-p",
            "tests.cli.test_server",
            "--schema",
            selector,
            "-o",
            str(tmp_path),
            str(query_file_path),
        ],
    )

    assert result.exit_code == 1
    assert "Error: Plugin tests.cli.test_server not found" in result.output


def test_codegen_finds_our_plugins(
    cli_app: Typer, cli_runner: CliRunner, query_file_path: Path, tmp_path: Path
):
    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "-p",
            "python",
            "--schema",
            selector,
            "-o",
            str(tmp_path),
            str(query_file_path),
        ],
    )

    assert result.exit_code == 0

    code_path = tmp_path / query_file_path.with_suffix(".py").name

    assert code_path.exists()
    assert "class GetUserResult" in code_path.read_text()


def test_can_use_custom_cli_plugin(
    cli_app: Typer, cli_runner: CliRunner, query_file_path: Path, tmp_path: Path
):
    selector = "tests.fixtures.sample_package.sample_module:schema"
    result = cli_runner.invoke(
        cli_app,
        [
            "codegen",
            "--cli-plugin",
            "tests.cli.test_codegen:ConsoleTestPlugin",
            "-p",
            "python",
            "--schema",
            selector,
            "-o",
            str(tmp_path),
            str(query_file_path),
        ],
    )

    assert result.exit_code == 0

    code_path = tmp_path / "renamed.py"

    assert code_path.exists()
    assert "class GetUserResult" in code_path.read_text()



import textwrap

import strawberry


def test_interface_object():
    @strawberry.federation.interface_object(keys=["id"])
    class SomeInterface:
        id: strawberry.ID

    schema = strawberry.federation.Schema(types=[SomeInterface])

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@interfaceObject", "@key"]) {
          query: Query
        }

        type Query {
          _entities(representations: [_Any!]!): [_Entity]!
          _service: _Service!
        }

        type SomeInterface @key(fields: "id") @interfaceObject {
          id: ID!
        }

        scalar _Any

        union _Entity = SomeInterface

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()



from __future__ import annotations

from collections.abc import AsyncIterable

from django.core.exceptions import BadRequest, SuspiciousOperation
from django.http import Http404, HttpRequest, HttpResponse, StreamingHttpResponse

from strawberry.django.views import AsyncGraphQLView as BaseAsyncGraphQLView
from strawberry.http import GraphQLHTTPResponse
from strawberry.http.ides import GraphQL_IDE
from strawberry.schema import Schema
from strawberry.types import ExecutionResult
from tests.http.context import get_context
from tests.views.schema import Query

from .base import Response, ResultOverrideFunction
from .django import DjangoHttpClient


class AsyncGraphQLView(BaseAsyncGraphQLView[dict[str, object], object]):
    result_override: ResultOverrideFunction = None

    async def get_root_value(self, request: HttpRequest) -> Query:
        await super().get_root_value(request)  # for coverage
        return Query()

    async def get_context(
        self, request: HttpRequest, response: HttpResponse
    ) -> dict[str, object]:
        context = {"request": request, "response": response}

        return get_context(context)

    async def process_result(
        self, request: HttpRequest, result: ExecutionResult
    ) -> GraphQLHTTPResponse:
        if self.result_override:
            return self.result_override(result)

        return await super().process_result(request, result)


class AsyncDjangoHttpClient(DjangoHttpClient):
    def __init__(
        self,
        schema: Schema,
        graphql_ide: GraphQL_IDE | None = "graphiql",
        allow_queries_via_get: bool = True,
        result_override: ResultOverrideFunction = None,
        multipart_uploads_enabled: bool = False,
    ):
        self.view = AsyncGraphQLView.as_view(
            schema=schema,
            graphql_ide=graphql_ide,
            allow_queries_via_get=allow_queries_via_get,
            result_override=result_override,
            multipart_uploads_enabled=multipart_uploads_enabled,
        )

    async def _do_request(self, request: HttpRequest) -> Response:
        try:
            response = await self.view(request)
        except Http404:
            return Response(status_code=404, data=b"Not found", headers={})
        except (BadRequest, SuspiciousOperation) as e:
            return Response(
                status_code=400,
                data=e.args[0].encode(),
                headers={},
            )

        data = (
            response.streaming_content
            if isinstance(response, StreamingHttpResponse)
            and isinstance(response.streaming_content, AsyncIterable)
            else response.content
        )

        return Response(
            status_code=response.status_code,
            data=data,
            headers=dict(response.headers),
        )



import textwrap
from enum import Enum
from typing import Annotated

import pytest

import strawberry


def test_field_inaccessible_printed_correctly():
    @strawberry.federation.interface(inaccessible=True)
    class AnInterface:
        id: strawberry.ID

    @strawberry.interface
    class SomeInterface:
        id: strawberry.ID
        a_field: str = strawberry.federation.field(inaccessible=True)

    @strawberry.federation.type(keys=["upc"], extend=True)
    class Product(SomeInterface):
        upc: str = strawberry.federation.field(external=True, inaccessible=True)

    @strawberry.federation.input(inaccessible=True)
    class AnInput:
        id: strawberry.ID = strawberry.federation.field(inaccessible=True)

    @strawberry.federation.type(inaccessible=True)
    class AnInaccessibleType:
        id: strawberry.ID

    @strawberry.federation.type
    class Query:
        @strawberry.field
        def top_products(
            self,
            first: Annotated[int, strawberry.federation.argument(inaccessible=True)],
        ) -> list[Product]:  # pragma: no cover
            return []

    schema = strawberry.federation.Schema(
        query=Query,
        types=[AnInterface, AnInput, AnInaccessibleType],
    )

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@external", "@inaccessible", "@key"]) {
          query: Query
        }

        type AnInaccessibleType @inaccessible {
          id: ID!
        }

        input AnInput @inaccessible {
          id: ID! @inaccessible
        }

        interface AnInterface @inaccessible {
          id: ID!
        }

        extend type Product implements SomeInterface @key(fields: "upc") {
          id: ID!
          aField: String! @inaccessible
          upc: String! @external @inaccessible
        }

        type Query {
          _entities(representations: [_Any!]!): [_Entity]!
          _service: _Service!
          topProducts(first: Int! @inaccessible): [Product!]!
        }

        interface SomeInterface {
          id: ID!
          aField: String! @inaccessible
        }

        scalar _Any

        union _Entity = Product

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_inaccessible_on_mutation():
    @strawberry.type
    class Query:
        hello: str

    @strawberry.type
    class Mutation:
        @strawberry.federation.mutation(inaccessible=True)
        def hello(self) -> str:  # pragma: no cover
            return "Hello"

    schema = strawberry.federation.Schema(
        query=Query,
        mutation=Mutation,
    )

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@inaccessible"]) {
          query: Query
          mutation: Mutation
        }

        type Mutation {
          hello: String! @inaccessible
        }

        type Query {
          _service: _Service!
          hello: String!
        }

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_inaccessible_on_scalar():
    SomeScalar = strawberry.federation.scalar(str, name="SomeScalar", inaccessible=True)

    @strawberry.type
    class Query:
        hello: SomeScalar

    schema = strawberry.federation.Schema(
        query=Query,
    )

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@inaccessible"]) {
          query: Query
        }

        type Query {
          _service: _Service!
          hello: SomeScalar!
        }

        scalar SomeScalar @inaccessible

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_inaccessible_on_enum():
    @strawberry.federation.enum(inaccessible=True)
    class SomeEnum(Enum):
        A = "A"

    @strawberry.type
    class Query:
        hello: SomeEnum

    schema = strawberry.federation.Schema(
        query=Query,
    )

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@inaccessible"]) {
          query: Query
        }

        type Query {
          _service: _Service!
          hello: SomeEnum!
        }

        enum SomeEnum @inaccessible {
          A
        }

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_inaccessible_on_enum_value():
    @strawberry.enum
    class SomeEnum(Enum):
        A = strawberry.federation.enum_value("A", inaccessible=True)

    @strawberry.type
    class Query:
        hello: SomeEnum

    schema = strawberry.federation.Schema(
        query=Query,
    )

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@inaccessible"]) {
          query: Query
        }

        type Query {
          _service: _Service!
          hello: SomeEnum!
        }

        enum SomeEnum {
          A @inaccessible
        }

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_field_tag_printed_correctly_on_union():
    @strawberry.type
    class A:
        a: str

    @strawberry.type
    class B:
        b: str

    MyUnion = Annotated[A | B, strawberry.federation.union("Union", inaccessible=True)]

    @strawberry.federation.type
    class Query:
        hello: MyUnion

    schema = strawberry.federation.Schema(query=Query)

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@inaccessible"]) {
          query: Query
        }

        type A {
          a: String!
        }

        type B {
          b: String!
        }

        type Query {
          _service: _Service!
          hello: Union!
        }

        union Union @inaccessible = A | B

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


@pytest.mark.parametrize("inaccessible", [True, False])
def test_inaccessible_on_type_and_interface(inaccessible: bool):
    """Test that inaccessible correctly adds/omits @inaccessible on types and interfaces."""

    @strawberry.federation.type(inaccessible=inaccessible)
    class SomeType:
        id: strawberry.ID

    @strawberry.federation.interface(inaccessible=inaccessible)
    class SomeInterface:
        id: strawberry.ID

    @strawberry.federation.type
    class Query:
        @strawberry.field
        def some(self) -> SomeType:  # pragma: no cover
            return SomeType(id=strawberry.ID("1"))

    schema = strawberry.federation.Schema(
        query=Query,
        types=[SomeInterface],
    )

    schema_str = schema.as_str()
    if inaccessible:
        assert "SomeType @inaccessible" in schema_str
        assert "SomeInterface @inaccessible" in schema_str
    else:
        assert "SomeType @inaccessible" not in schema_str
        assert "SomeInterface @inaccessible" not in schema_str



import dataclasses
from textwrap import dedent

import strawberry
from strawberry.types.unset import UNSET


def test_mutation():
    @strawberry.type
    class Query:
        hello: str = "Hello"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def say(self) -> str:
            return "Hello!"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    query = "mutation { say }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["say"] == "Hello!"


def test_mutation_with_input_type():
    @strawberry.input
    class SayInput:
        name: str
        age: int

    @strawberry.type
    class Query:
        hello: str = "Hello"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def say(self, input: SayInput) -> str:
            return f"Hello {input.name} of {input.age} years old!"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    query = 'mutation { say(input: { name: "Patrick", age: 10 }) }'

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["say"] == "Hello Patrick of 10 years old!"


def test_mutation_reusing_input_types():
    @strawberry.input
    class SayInput:
        name: str
        age: int

    @strawberry.type
    class Query:
        hello: str = "Hello"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def say(self, input: SayInput) -> str:
            return f"Hello {input.name} of {input.age} years old!"

        @strawberry.mutation
        def say2(self, input: SayInput) -> str:
            return f"Hello {input.name} of {input.age}!"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    query = 'mutation { say2(input: { name: "Patrick", age: 10 }) }'

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["say2"] == "Hello Patrick of 10!"


def test_unset_types():
    @strawberry.type
    class Query:
        hello: str = "Hello"

    @strawberry.input
    class InputExample:
        name: str
        age: int | None = UNSET

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def say(self, name: str | None = UNSET) -> str:  # type: ignore
            if name is UNSET:
                return "Name is unset"

            return f"Hello {name}!"

        @strawberry.mutation
        def say_age(self, input: InputExample) -> str:
            age = "unset" if input.age is UNSET else input.age

            return f"Hello {input.name} of age {age}!"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    query = 'mutation { say sayAge(input: { name: "P"}) }'

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["say"] == "Name is unset"
    assert result.data["sayAge"] == "Hello P of age unset!"


def test_unset_types_name_with_underscore():
    @strawberry.type
    class Query:
        hello: str = "Hello"

    @strawberry.input
    class InputExample:
        first_name: str
        age: str | None = UNSET

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def say(self, first_name: str | None = UNSET) -> str:  # type: ignore
            if first_name is UNSET:
                return "Name is unset"

            if first_name == "":
                return "Hello Empty!"

            return f"Hello {first_name}!"

        @strawberry.mutation
        def say_age(self, input: InputExample) -> str:
            age = "unset" if input.age is UNSET else input.age
            age = "empty" if age == "" else age

            return f"Hello {input.first_name} of age {age}!"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    query = """mutation {
        one: say
        two: say(firstName: "Patrick")
        three: say(firstName: "")
        empty: sayAge(input: { firstName: "Patrick", age: "" })
        null: sayAge(input: { firstName: "Patrick", age: null })
        sayAge(input: { firstName: "Patrick" })
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["one"] == "Name is unset"
    assert result.data["two"] == "Hello Patrick!"
    assert result.data["three"] == "Hello Empty!"
    assert result.data["empty"] == "Hello Patrick of age empty!"
    assert result.data["null"] == "Hello Patrick of age None!"
    assert result.data["sayAge"] == "Hello Patrick of age unset!"


def test_unset_types_stringify_empty():
    @strawberry.type
    class Query:
        hello: str = "Hello"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def say(self, first_name: str | None = UNSET) -> str:  # type: ignore
            return f"Hello {first_name}!"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    query = """mutation {
        say
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["say"] == "Hello !"

    query = """mutation {
        say(firstName: null)
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["say"] == "Hello None!"


def test_converting_to_dict_with_unset():
    @strawberry.type
    class Query:
        hello: str = "Hello"

    @strawberry.input
    class Input:
        name: str | None = UNSET

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def say(self, input: Input) -> str:
            data = dataclasses.asdict(input)

            if data["name"] is UNSET:
                return "Hello 🤨"

            return f"Hello {data['name']}!"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    query = """mutation {
        say(input: {})
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["say"] == "Hello 🤨"


def test_mutation_deprecation_reason():
    @strawberry.type
    class Query:
        hello: str = "world"

    @strawberry.type
    class Mutation:
        @strawberry.mutation(deprecation_reason="Your reason")
        def say(self, name: str) -> str:
            return f"Hello {name}!"

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    assert str(schema) == dedent(
        """\
        type Mutation {
          say(name: String!): String! @deprecated(reason: "Your reason")
        }

        type Query {
          hello: String!
        }"""
    )



from __future__ import annotations

import os
import pathlib
import re
import subprocess
import tempfile
from typing import cast

from .result import Result, ResultType


def run_ty(code: str, strict: bool = True) -> list[Result]:
    # ty uses concise output format which includes revealed type info
    # Format: <file>:<line>:<col>: <severity>[<check_name>] <message>
    args = [
        "ty",
        "check",
        "--output-format",
        "concise",
        # Ignore the warning about using reveal_type without importing it
        "--ignore",
        "undefined-reveal",
        "--color",
        "never",
    ]

    with tempfile.TemporaryDirectory() as directory:
        module_path = pathlib.Path(directory) / "ty_test.py"
        module_path.write_text(code)

        process_result = subprocess.run(
            [*args, str(module_path)],
            check=False,
            capture_output=True,
            env={
                "PATH": os.environ["PATH"],
            },
        )

        full_output = (
            process_result.stdout.decode("utf-8")
            + "\n"
            + process_result.stderr.decode("utf-8")
        )
        full_output = full_output.strip()

        results: list[Result] = []

        # Parse the concise output format
        # Format: <file>:<line>:<col>: <severity>[<check_name>] <message>
        pattern = re.compile(r"^.*?:(\d+):(\d+): (error|warning|info)\[[\w-]+\] (.+)$")

        for raw_line in full_output.split("\n"):
            line = raw_line.strip()
            if not line:
                continue

            match = pattern.match(line)
            if match:
                line_num = int(match.group(1))
                col_num = int(match.group(2))
                severity = match.group(3)
                message = match.group(4)

                # Map ty severities to our ResultType
                # ty uses: error, warning, info
                # Our ResultType uses: error, information, note
                type_mapping = {
                    "error": "error",
                    "warning": "error",  # treat warnings as errors for consistency
                    "info": "information",
                }
                result_type = type_mapping.get(severity, "note")

                results.append(
                    Result(
                        type=cast("ResultType", result_type),
                        message=message.strip(),
                        line=line_num,
                        column=col_num,
                    )
                )

        # Sort results by line, column, and message for consistent ordering
        results.sort(key=lambda x: (x.line, x.column, x.message))

        return results



from typing import Annotated, Any, cast, get_args

import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.types.auto import StrawberryAuto, auto
from strawberry.types.base import StrawberryList


@strawberry.type
class ExampleType:
    some_var: str


def test_singleton():
    assert get_args(auto)[1] is StrawberryAuto()
    assert StrawberryAuto() is StrawberryAuto()


def test_annotated():
    assert get_args(auto) == (Any, StrawberryAuto())
    some_obj = object()
    new_annotated = Annotated[strawberry.auto, some_obj]
    assert get_args(new_annotated) == (Any, StrawberryAuto(), some_obj)


def test_str():
    assert str(StrawberryAuto()) == "auto"


def test_repr():
    assert repr(StrawberryAuto()) == "<auto>"


def test_isinstance():
    assert isinstance(auto, StrawberryAuto)
    assert not isinstance(object, StrawberryAuto)
    assert not isinstance(cast("Any", object()), StrawberryAuto)


def test_isinstance_with_annotation():
    annotation = StrawberryAnnotation(auto)
    assert isinstance(annotation, StrawberryAuto)
    str_annotation = StrawberryAnnotation("auto", namespace=globals())
    assert isinstance(str_annotation, StrawberryAuto)


def test_isinstance_with_annotated():
    assert isinstance(Annotated[auto, object()], StrawberryAuto)
    assert not isinstance(Annotated[str, strawberry.auto], StrawberryAuto)


def test_isinstance_with_unresolvable_annotation():
    type_ = StrawberryList(of_type=ExampleType)
    assert not isinstance(type_, StrawberryAuto)



from typing import Any

from fastapi import BackgroundTasks, Depends, FastAPI, Request, WebSocket

from strawberry.fastapi import GraphQLRouter
from tests.views.schema import schema


def custom_context_dependency() -> str:
    return "Hi!"


async def get_context(
    background_tasks: BackgroundTasks,
    request: Request = None,
    ws: WebSocket = None,
    custom_value=Depends(custom_context_dependency),
) -> dict[str, Any]:
    return {
        "custom_value": custom_value,
        "request": request or ws,
        "background_tasks": background_tasks,
    }


async def get_root_value(
    request: Request = None, ws: WebSocket = None
) -> Request | WebSocket:
    return request or ws


def create_app(schema=schema, **kwargs: Any) -> FastAPI:
    app = FastAPI()

    graphql_app = GraphQLRouter(
        schema, context_getter=get_context, root_value_getter=get_root_value, **kwargs
    )
    app.include_router(graphql_app, prefix="/graphql")

    return app



import pytest

import strawberry

pytestmark = pytest.mark.pydantic


def test_use_alias_as_gql_name():
    from pydantic import BaseModel, Field

    class UserModel(BaseModel):
        age_: int = Field(..., alias="age_alias")

    @strawberry.experimental.pydantic.type(
        UserModel, all_fields=True, use_pydantic_alias=True
    )
    class User: ...

    @strawberry.type
    class Query:
        user: User = strawberry.field(default_factory=lambda: User(age_=5))

    schema = strawberry.Schema(query=Query)
    query = """{
        user {
            __typename,

            ... on User {
                age_alias
            }
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["user"] == {"__typename": "User", "age_alias": 5}


def test_do_not_use_alias_as_gql_name():
    from pydantic import BaseModel, Field

    class UserModel(BaseModel):
        age_: int = Field(..., alias="age_alias")

    @strawberry.experimental.pydantic.type(
        UserModel, all_fields=True, use_pydantic_alias=False
    )
    class User: ...

    @strawberry.type
    class Query:
        user: User = strawberry.field(default_factory=lambda: User(age_=5))

    schema = strawberry.Schema(query=Query)
    query = """{
        user {
            __typename,

            ... on User {
                age_
            }
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["user"] == {"__typename": "User", "age_": 5}



from io import BytesIO

from strawberry.file_uploads.utils import replace_placeholders_with_files


def test_does_deep_copy():
    operations = {
        "query": "mutation($file: Upload!) { upload_file(file: $file) { id } }",
        "variables": {"file": None},
    }
    files_map = {}
    files = {}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result == operations
    assert result is not operations


def test_empty_files_map():
    operations = {
        "query": "mutation($files: [Upload!]!) { upload_files(files: $files) { id } }",
        "variables": {"files": [None, None]},
    }
    files_map = {}
    files = {"0": BytesIO(), "1": BytesIO()}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result == operations


def test_empty_operations_paths():
    operations = {
        "query": "mutation($files: [Upload!]!) { upload_files(files: $files) { id } }",
        "variables": {"files": [None, None]},
    }
    files_map = {"0": [], "1": []}
    files = {"0": BytesIO(), "1": BytesIO()}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result == operations


def test_single_file_in_single_location():
    operations = {
        "query": "mutation($file: Upload!) { upload_file(file: $file) { id } }",
        "variables": {"file": None},
    }
    files_map = {"0": ["variables.file"]}
    file0 = BytesIO()
    files = {"0": file0}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result["query"] == operations["query"]
    assert result["variables"]["file"] == file0


def test_single_file_in_multiple_locations():
    operations = {
        "query": "mutation($a: Upload!, $b: Upload!) { pair(a: $a, b: $a) { id } }",
        "variables": {"a": None, "b": None},
    }
    files_map = {"0": ["variables.a", "variables.b"]}
    file0 = BytesIO()
    files = {"0": file0}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result["query"] == operations["query"]
    assert result["variables"]["a"] == file0
    assert result["variables"]["b"] == file0


def test_file_list():
    operations = {
        "query": "mutation($files: [Upload!]!) { upload_files(files: $files) { id } }",
        "variables": {"files": [None, None]},
    }
    files_map = {"0": ["variables.files.0"], "1": ["variables.files.1"]}
    file0 = BytesIO()
    file1 = BytesIO()
    files = {"0": file0, "1": file1}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result["query"] == operations["query"]
    assert result["variables"]["files"][0] == file0
    assert result["variables"]["files"][1] == file1


def test_single_file_reuse_in_list():
    operations = {
        "query": "mutation($a: [Upload!]!, $b: Upload!) { mixed(a: $a, b: $b) { id } }",
        "variables": {"a": [None, None], "b": None},
    }
    files_map = {"0": ["variables.a.0"], "1": ["variables.a.1", "variables.b"]}
    file0 = BytesIO()
    file1 = BytesIO()
    files = {"0": file0, "1": file1}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result["query"] == operations["query"]
    assert result["variables"]["a"][0] == file0
    assert result["variables"]["a"][1] == file1
    assert result["variables"]["b"] == file1


def test_using_single_file_multiple_times_in_same_list():
    operations = {
        "query": "mutation($files: [Upload!]!) { upload_files(files: $files) { id } }",
        "variables": {"files": [None, None]},
    }
    files_map = {"0": ["variables.files.0", "variables.files.1"]}
    file0 = BytesIO()
    files = {"0": file0}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result["query"] == operations["query"]
    assert result["variables"]["files"][0] == file0
    assert result["variables"]["files"][1] == file0


def test_deep_nesting():
    operations = {
        "query": "mutation($list: [ComplexInput!]!) { mutate(list: $list) { id } }",
        "variables": {"a": [{"files": [None, None]}]},
    }
    files_map = {"0": ["variables.a.0.files.0"], "1": ["variables.a.0.files.1"]}
    file0 = BytesIO()
    file1 = BytesIO()
    files = {"0": file0, "1": file1}

    result = replace_placeholders_with_files(operations, files_map, files)
    assert result["query"] == operations["query"]
    assert result["variables"]["a"][0]["files"][0] == file0
    assert result["variables"]["a"][0]["files"][1] == file1



import asyncio
from collections.abc import AsyncGenerator

import pytest

import strawberry
from strawberry.exceptions import InvalidCustomContext
from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL
from strawberry.subscriptions.protocols.graphql_transport_ws import (
    types as transport_ws_types,
)
from strawberry.subscriptions.protocols.graphql_ws import types as ws_types


def test_base_context():
    from strawberry.fastapi import BaseContext

    base_context = BaseContext()
    assert base_context.request is None
    assert base_context.background_tasks is None
    assert base_context.response is None


def test_with_explicit_class_context_getter():
    from fastapi import Depends, FastAPI
    from fastapi.testclient import TestClient

    from strawberry.fastapi import BaseContext, GraphQLRouter

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert info.context.request is not None
            assert info.context.strawberry == "explicitly rocks"
            assert info.context.connection_params is None
            return "abc"

    class CustomContext(BaseContext):
        def __init__(self, rocks: str):
            self.strawberry = rocks

    def custom_context_dependency() -> CustomContext:
        return CustomContext(rocks="explicitly rocks")

    def get_context(custom_context: CustomContext = Depends(custom_context_dependency)):
        return custom_context

    app = FastAPI()
    schema = strawberry.Schema(query=Query)
    graphql_app = GraphQLRouter(schema=schema, context_getter=get_context)
    app.include_router(graphql_app, prefix="/graphql")

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"abc": "abc"}}


def test_with_implicit_class_context_getter():
    from fastapi import Depends, FastAPI
    from fastapi.testclient import TestClient

    from strawberry.fastapi import BaseContext, GraphQLRouter

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert info.context.request is not None
            assert info.context.strawberry == "implicitly rocks"
            assert info.context.connection_params is None
            return "abc"

    class CustomContext(BaseContext):
        def __init__(self, rocks: str = "implicitly rocks"):
            super().__init__()
            self.strawberry = rocks

    def get_context(custom_context: CustomContext = Depends()):
        return custom_context

    app = FastAPI()
    schema = strawberry.Schema(query=Query)
    graphql_app = GraphQLRouter(schema=schema, context_getter=get_context)
    app.include_router(graphql_app, prefix="/graphql")

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"abc": "abc"}}


def test_with_dict_context_getter():
    from fastapi import Depends, FastAPI
    from fastapi.testclient import TestClient

    from strawberry.fastapi import GraphQLRouter

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert info.context.get("request") is not None
            assert "connection_params" not in info.context
            assert info.context.get("strawberry") == "rocks"
            return "abc"

    def custom_context_dependency() -> str:
        return "rocks"

    def get_context(value: str = Depends(custom_context_dependency)) -> dict[str, str]:
        return {"strawberry": value}

    app = FastAPI()
    schema = strawberry.Schema(query=Query)
    graphql_app = GraphQLRouter(schema=schema, context_getter=get_context)
    app.include_router(graphql_app, prefix="/graphql")

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"abc": "abc"}}


def test_without_context_getter():
    from fastapi import FastAPI
    from fastapi.testclient import TestClient

    from strawberry.fastapi import GraphQLRouter

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert info.context.get("request") is not None
            assert info.context.get("strawberry") is None
            return "abc"

    app = FastAPI()
    schema = strawberry.Schema(query=Query)
    graphql_app = GraphQLRouter(schema, context_getter=None)
    app.include_router(graphql_app, prefix="/graphql")

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"abc": "abc"}}


def test_with_invalid_context_getter():
    from fastapi import Depends, FastAPI
    from fastapi.testclient import TestClient

    from strawberry.fastapi import GraphQLRouter

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert info.context.get("request") is not None
            assert info.context.get("strawberry") is None
            return "abc"

    def custom_context_dependency() -> str:
        return "rocks"

    def get_context(value: str = Depends(custom_context_dependency)) -> str:
        return value

    app = FastAPI()
    schema = strawberry.Schema(query=Query)
    graphql_app = GraphQLRouter(schema=schema, context_getter=get_context)
    app.include_router(graphql_app, prefix="/graphql")

    test_client = TestClient(app)
    with pytest.raises(
        InvalidCustomContext,
        match=(
            "The custom context must be either a class "
            "that inherits from BaseContext or a dictionary"
        ),
    ):
        test_client.post("/graphql", json={"query": "{ abc }"})


def test_class_context_injects_connection_params_over_transport_ws():
    from fastapi import Depends, FastAPI
    from fastapi.testclient import TestClient

    from strawberry.fastapi import BaseContext, GraphQLRouter

    @strawberry.type
    class Query:
        x: str = "hi"

    @strawberry.type
    class Subscription:
        @strawberry.subscription
        async def connection_params(
            self, info: strawberry.Info, delay: float = 0
        ) -> AsyncGenerator[str, None]:
            assert info.context.request is not None
            await asyncio.sleep(delay)
            yield info.context.connection_params["strawberry"]

    class Context(BaseContext):
        strawberry: str

        def __init__(self):
            self.strawberry = "rocks"

    def get_context(context: Context = Depends()) -> Context:
        return context

    app = FastAPI()
    schema = strawberry.Schema(query=Query, subscription=Subscription)
    graphql_app = GraphQLRouter(schema=schema, context_getter=get_context)
    app.include_router(graphql_app, prefix="/graphql")
    test_client = TestClient(app)

    with test_client.websocket_connect(
        "/graphql", [GRAPHQL_TRANSPORT_WS_PROTOCOL]
    ) as ws:
        ws.send_json(
            transport_ws_types.ConnectionInitMessage(
                {"type": "connection_init", "payload": {"strawberry": "rocks"}}
            )
        )

        connection_ack_message: transport_ws_types.ConnectionInitMessage = (
            ws.receive_json()
        )
        assert connection_ack_message == {"type": "connection_ack"}

        ws.send_json(
            transport_ws_types.SubscribeMessage(
                {
                    "id": "sub1",
                    "type": "subscribe",
                    "payload": {"query": "subscription { connectionParams }"},
                }
            )
        )

        next_message: transport_ws_types.NextMessage = ws.receive_json()
        assert next_message == {
            "id": "sub1",
            "type": "next",
            "payload": {"data": {"connectionParams": "rocks"}},
        }

        ws.send_json(
            transport_ws_types.CompleteMessage({"id": "sub1", "type": "complete"})
        )

        ws.close()


def test_class_context_injects_connection_params_over_ws():
    from fastapi import Depends, FastAPI
    from fastapi.testclient import TestClient
    from starlette.websockets import WebSocketDisconnect

    from strawberry.fastapi import BaseContext, GraphQLRouter

    @strawberry.type
    class Query:
        x: str = "hi"

    @strawberry.type
    class Subscription:
        @strawberry.subscription
        async def connection_params(
            self, info: strawberry.Info, delay: float = 0
        ) -> AsyncGenerator[str, None]:
            assert info.context.request is not None
            await asyncio.sleep(delay)
            yield info.context.connection_params["strawberry"]

    class Context(BaseContext):
        strawberry: str

        def __init__(self):
            self.strawberry = "rocks"

    def get_context(context: Context = Depends()) -> Context:
        return context

    app = FastAPI()
    schema = strawberry.Schema(query=Query, subscription=Subscription)
    graphql_app = GraphQLRouter(schema=schema, context_getter=get_context)
    app.include_router(graphql_app, prefix="/graphql")
    test_client = TestClient(app)

    with test_client.websocket_connect("/graphql", [GRAPHQL_WS_PROTOCOL]) as ws:
        ws.send_json(
            ws_types.ConnectionInitMessage(
                {
                    "type": "connection_init",
                    "payload": {"strawberry": "rocks"},
                }
            )
        )
        ws.send_json(
            ws_types.StartMessage(
                {
                    "type": "start",
                    "id": "demo",
                    "payload": {
                        "query": "subscription { connectionParams }",
                    },
                }
            )
        )

        connection_ack_message: ws_types.ConnectionAckMessage = ws.receive_json()
        assert connection_ack_message["type"] == "connection_ack"

        data_message: ws_types.DataMessage = ws.receive_json()
        assert data_message["type"] == "data"
        assert data_message["id"] == "demo"
        assert data_message["payload"]["data"] == {"connectionParams": "rocks"}

        ws.send_json(ws_types.StopMessage({"type": "stop", "id": "demo"}))

        complete_message: ws_types.CompleteMessage = ws.receive_json()
        assert complete_message["type"] == "complete"
        assert complete_message["id"] == "demo"

        ws.send_json(
            ws_types.ConnectionTerminateMessage({"type": "connection_terminate"})
        )

        # make sure the websocket is disconnected now
        with pytest.raises(WebSocketDisconnect):
            ws.receive_json()



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, requires_ty, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy, requires_ty]


CODE = """
import strawberry
from strawberry.scalars import ID, JSON, Base16, Base32, Base64


@strawberry.type
class SomeType:
    id: ID
    json: JSON
    base16: Base16
    base32: Base32
    base64: Base64


obj = SomeType(
    id=ID("123"),
    json=JSON({"foo": "bar"}),
    base16=Base16(b"<bytes>"),
    base32=Base32(b"<bytes>"),
    base64=Base64(b"<bytes>"),
)

reveal_type(obj.id)
reveal_type(obj.json)
reveal_type(obj.base16)
reveal_type(obj.base16)
reveal_type(obj.base64)
"""


def test():
    results = typecheck(CODE)

    # NOTE: This is also guaranteeing that those scalars could be used to annotate
    # the attributes. Pyright 1.1.224+ doesn't allow non-types to be used there
    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "obj.id" is "ID"',
                line=23,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "obj.json" is "JSON"',
                line=24,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "obj.base16" is "Base16"',
                line=25,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "obj.base16" is "Base16"',
                line=26,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "obj.base64" is "Base64"',
                line=27,
                column=13,
            ),
        ]
    )
    # Now that scalars are proper NewTypes, mypy correctly identifies them
    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "strawberry.scalars.ID"',
                line=23,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.scalars.JSON"',
                line=24,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.scalars.Base16"',
                line=25,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.scalars.Base16"',
                line=26,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.scalars.Base64"',
                line=27,
                column=13,
            ),
        ]
    )
    # Now that scalars are proper NewTypes, ty no longer reports errors
    assert results.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `ID`",
                line=23,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `JSON`",
                line=24,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `Base16`",
                line=25,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `Base16`",
                line=26,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `Base64`",
                line=27,
                column=13,
            ),
        ]
    )


CODE_SCHEMA_OVERRIDES = """
import strawberry
from datetime import datetime, timezone

EpochDateTime = strawberry.scalar(
    datetime,
)

@strawberry.type
class Query:
    a: datetime

schema = strawberry.Schema(query=Query, scalar_overrides={
    datetime: EpochDateTime,
})

reveal_type(EpochDateTime)
"""


def test_schema_overrides():
    # TODO: change strict to true when we improve type hints for scalar
    results = typecheck(CODE_SCHEMA_OVERRIDES, strict=False)

    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "EpochDateTime" is "type[datetime]"',
                line=16,
                column=13,
            )
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "def (year: typing.SupportsIndex, month: typing.SupportsIndex, day: typing.SupportsIndex, hour: typing.SupportsIndex =, minute: typing.SupportsIndex =, second: typing.SupportsIndex =, microsecond: typing.SupportsIndex =, tzinfo: datetime.tzinfo | None =, *, fold: int =) -> datetime.datetime"',
                line=17,
                column=13,
            )
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `<class 'datetime'>`",
                line=17,
                column=13,
            ),
        ]
    )


CODE_SCALAR_MAP = """
import strawberry
from typing import NewType
from strawberry.schema.config import StrawberryConfig

MyString = NewType("MyString", str)

@strawberry.type
class Query:
    value: MyString

schema = strawberry.Schema(
    query=Query,
    config=StrawberryConfig(
        scalar_map={
            MyString: strawberry.scalar(
                name="MyString",
                serialize=lambda v: v.upper(),
            )
        }
    ),
)

reveal_type(MyString)
reveal_type(MyString("test"))
"""


def test_scalar_map():
    results = typecheck(CODE_SCALAR_MAP, strict=False)

    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "MyString" is "type[MyString]"',
                line=23,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "MyString("test")" is "MyString"',
                line=24,
                column=13,
            ),
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "def (item: str) -> mypy_test.MyString"',
                line=24,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "mypy_test.MyString"',
                line=25,
                column=13,
            ),
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `<NewType pseudo-class 'MyString'>`",
                line=24,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `MyString`",
                line=25,
                column=13,
            ),
        ]
    )


CODE_NEWTYPE_SCALAR_USAGE = """
import strawberry
from strawberry.scalars import JSON, Base64

@strawberry.type
class Query:
    # Test using NewType scalars in various positions
    json_field: JSON
    base64_field: Base64

    @strawberry.field
    def process_json(self, data: JSON) -> JSON:
        return data

    @strawberry.field
    def encode_data(self, data: str) -> Base64:
        return Base64(data.encode())

# Test that NewType scalars can be instantiated correctly
json_val: JSON = JSON({"key": "value"})
base64_val: Base64 = Base64(b"hello")

reveal_type(json_val)
reveal_type(base64_val)
"""


def test_newtype_scalar_usage():
    results = typecheck(CODE_NEWTYPE_SCALAR_USAGE, strict=False)

    # Verify pyright sees the correct types
    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "json_val" is "JSON"',
                line=22,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "base64_val" is "Base64"',
                line=23,
                column=13,
            ),
        ]
    )
    # Verify mypy sees the correct types
    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "strawberry.scalars.JSON"',
                line=23,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.scalars.Base64"',
                line=24,
                column=13,
            ),
        ]
    )
    # Verify ty sees the correct types (no errors about invalid type expressions)
    assert results.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `JSON`",
                line=23,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `Base64`",
                line=24,
                column=13,
            ),
        ]
    )



import pathlib
import textwrap

from pytest_snapshot.plugin import Snapshot

import strawberry
from strawberry import relay
from strawberry.relay.utils import to_base64

from .schema import schema
from .schema_future_annotations import schema as schema_future_annotations

SNAPSHOTS_DIR = pathlib.Path(__file__).parent / "snapshots"


def test_schema(snapshot: Snapshot):
    snapshot.snapshot_dir = SNAPSHOTS_DIR
    snapshot.assert_match(str(schema), "schema.gql")


def test_schema_future_annotations(snapshot: Snapshot):
    snapshot.snapshot_dir = SNAPSHOTS_DIR
    snapshot.assert_match(
        str(schema_future_annotations),
        "schema_future_annotations.gql",
    )


def test_node_id_annotation():
    @strawberry.type
    class Fruit(relay.Node):
        code: relay.NodeID[int]

    @strawberry.type
    class Query:
        @relay.connection(relay.ListConnection[Fruit])
        def fruits(self) -> list[Fruit]:
            return [Fruit(code=i) for i in range(10)]

    schema = strawberry.Schema(query=Query)
    expected = textwrap.dedent(
        '''
        type Fruit implements Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        """A connection to a list of items."""
        type FruitConnection {
          """Pagination data for this connection"""
          pageInfo: PageInfo!

          """Contains the nodes in this connection"""
          edges: [FruitEdge!]!
        }

        """An edge in a connection."""
        type FruitEdge {
          """A cursor for use in pagination"""
          cursor: String!

          """The item at the end of the edge"""
          node: Fruit!
        }

        """An object with a Globally Unique ID"""
        interface Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        """Information to aid in pagination."""
        type PageInfo {
          """When paginating forwards, are there more items?"""
          hasNextPage: Boolean!

          """When paginating backwards, are there more items?"""
          hasPreviousPage: Boolean!

          """When paginating backwards, the cursor to continue."""
          startCursor: String

          """When paginating forwards, the cursor to continue."""
          endCursor: String
        }

        type Query {
          fruits(
            """Returns the items in the list that come before the specified cursor."""
            before: String = null

            """Returns the items in the list that come after the specified cursor."""
            after: String = null

            """Returns the first n items from the list."""
            first: Int = null

            """Returns the items in the list that come after the specified cursor."""
            last: Int = null
          ): FruitConnection!
        }
        '''
    ).strip()
    assert str(schema) == expected
    result = schema.execute_sync(
        """
        query {
          fruits {
            edges {
              node {
                id
              }
            }
          }
        }
        """
    )
    assert result.errors is None
    assert result.data == {
        "fruits": {
            "edges": [{"node": {"id": to_base64("Fruit", i)}} for i in range(10)]
        }
    }


def test_node_id_annotation_in_superclass():
    @strawberry.type
    class BaseFruit(relay.Node):
        code: relay.NodeID[int]

    @strawberry.type
    class Fruit(BaseFruit): ...

    @strawberry.type
    class Query:
        @relay.connection(relay.ListConnection[Fruit])
        def fruits(self) -> list[Fruit]:
            return [Fruit(code=i) for i in range(10)]

    schema = strawberry.Schema(query=Query)
    expected = textwrap.dedent(
        '''
        type Fruit implements Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        """A connection to a list of items."""
        type FruitConnection {
          """Pagination data for this connection"""
          pageInfo: PageInfo!

          """Contains the nodes in this connection"""
          edges: [FruitEdge!]!
        }

        """An edge in a connection."""
        type FruitEdge {
          """A cursor for use in pagination"""
          cursor: String!

          """The item at the end of the edge"""
          node: Fruit!
        }

        """An object with a Globally Unique ID"""
        interface Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        """Information to aid in pagination."""
        type PageInfo {
          """When paginating forwards, are there more items?"""
          hasNextPage: Boolean!

          """When paginating backwards, are there more items?"""
          hasPreviousPage: Boolean!

          """When paginating backwards, the cursor to continue."""
          startCursor: String

          """When paginating forwards, the cursor to continue."""
          endCursor: String
        }

        type Query {
          fruits(
            """Returns the items in the list that come before the specified cursor."""
            before: String = null

            """Returns the items in the list that come after the specified cursor."""
            after: String = null

            """Returns the first n items from the list."""
            first: Int = null

            """Returns the items in the list that come after the specified cursor."""
            last: Int = null
          ): FruitConnection!
        }
        '''
    ).strip()
    assert str(schema) == expected
    result = schema.execute_sync(
        """
        query {
          fruits {
            edges {
              node {
                id
              }
            }
          }
        }
        """
    )
    assert result.errors is None
    assert result.data == {
        "fruits": {
            "edges": [{"node": {"id": to_base64("Fruit", i)}} for i in range(10)]
        }
    }


def test_node_id_annotation_in_superclass_and_subclass():
    @strawberry.type
    class BaseFruit(relay.Node):
        code: relay.NodeID[int]

    @strawberry.type
    class Fruit(BaseFruit):
        other_code: relay.NodeID[int]

    @strawberry.type
    class Query:
        @relay.connection(relay.ListConnection[Fruit])
        def fruits(self) -> list[Fruit]:
            return [Fruit(code=i, other_code=i) for i in range(10)]

    schema = strawberry.Schema(query=Query)
    expected = textwrap.dedent(
        '''
        type Fruit implements Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        """A connection to a list of items."""
        type FruitConnection {
          """Pagination data for this connection"""
          pageInfo: PageInfo!

          """Contains the nodes in this connection"""
          edges: [FruitEdge!]!
        }

        """An edge in a connection."""
        type FruitEdge {
          """A cursor for use in pagination"""
          cursor: String!

          """The item at the end of the edge"""
          node: Fruit!
        }

        """An object with a Globally Unique ID"""
        interface Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        """Information to aid in pagination."""
        type PageInfo {
          """When paginating forwards, are there more items?"""
          hasNextPage: Boolean!

          """When paginating backwards, are there more items?"""
          hasPreviousPage: Boolean!

          """When paginating backwards, the cursor to continue."""
          startCursor: String

          """When paginating forwards, the cursor to continue."""
          endCursor: String
        }

        type Query {
          fruits(
            """Returns the items in the list that come before the specified cursor."""
            before: String = null

            """Returns the items in the list that come after the specified cursor."""
            after: String = null

            """Returns the first n items from the list."""
            first: Int = null

            """Returns the items in the list that come after the specified cursor."""
            last: Int = null
          ): FruitConnection!
        }
        '''
    ).strip()
    assert str(schema) == expected
    result = schema.execute_sync(
        """
        query {
          fruits {
            edges {
              node {
                id
              }
            }
          }
        }
        """
    )
    assert result.errors is None
    assert result.data == {
        "fruits": {
            "edges": [{"node": {"id": to_base64("Fruit", i)}} for i in range(10)]
        }
    }


def test_overwrite_resolve_id_and_no_node_id():
    @strawberry.type
    class Fruit(relay.Node):
        color: str

        @classmethod
        def resolve_id(cls, root) -> str:
            return "test"  # pragma: no cover

    @strawberry.type
    class Query:
        fruit: Fruit

    expected_type = textwrap.dedent(
        '''
          type Fruit implements Node {
          """The Globally Unique ID of this object"""
          id: ID!
          color: String!
        }

        """An object with a Globally Unique ID"""
        interface Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        type Query {
          fruit: Fruit!
        }
        '''
    )

    schema = strawberry.Schema(query=Query)

    assert str(schema) == textwrap.dedent(expected_type).strip()



from collections.abc import Iterable
from typing import Annotated, Any, Optional
from typing_extensions import Self

import pytest

import strawberry
from strawberry.permission import BasePermission
from strawberry.relay import Connection, Node, PageInfo, to_base64
from strawberry.relay.types import Edge, ListConnection
from strawberry.schema.config import StrawberryConfig


@strawberry.type
class User(Node):
    id: strawberry.relay.NodeID
    name: str = "John"

    @classmethod
    def resolve_nodes(
        cls, *, info: strawberry.Info, node_ids: list[Any], required: bool
    ) -> list[Self]:
        return [cls() for _ in node_ids]


@strawberry.type
class EmptyUserConnection(Connection[User]):
    @classmethod
    def resolve_connection(
        cls,
        nodes: Iterable[User],
        *,
        info: Any,
        after: str | None = None,
        before: str | None = None,
        first: int | None = None,
        last: int | None = None,
        max_results: int | None = None,
        **kwargs: Any,
    ) -> Self | None:
        return None


@strawberry.type
class UserConnection(Connection[User]):
    @classmethod
    def resolve_connection(
        cls,
        nodes: Iterable[User],
        *,
        info: Any,
        after: str | None = None,
        before: str | None = None,
        first: int | None = None,
        last: int | None = None,
        max_results: int | None = None,
        **kwargs: Any,
    ) -> Self | None:
        user_node_id = to_base64(User, "1")
        return cls(
            page_info=PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor=None,
                end_cursor=None,
            ),
            edges=[Edge(cursor=user_node_id, node=User(id=user_node_id))],
        )


class TestPermission(BasePermission):
    message = "Not allowed"

    def has_permission(self, source, info, **kwargs: Any):
        return False


def test_nullable_connection_with_optional():
    @strawberry.type
    class Query:
        @strawberry.relay.connection(Optional[EmptyUserConnection])
        def users(self) -> list[User] | None:
            return None

    schema = strawberry.Schema(query=Query)
    query = """
        query {
            users {
                edges {
                    node {
                        name
                    }
                }
            }
        }
    """

    result = schema.execute_sync(query)
    assert result.data == {"users": None}
    assert not result.errors


def test_lazy_connection():
    @strawberry.type
    class Query:
        @strawberry.relay.connection(
            Optional[
                Annotated[
                    "UserConnection", strawberry.lazy("tests.relay.test_connection")
                ]
            ]
        )
        def users(self) -> list[User] | None:
            return None

    schema = strawberry.Schema(query=Query)
    query = """
            query {
                users {
                    edges {
                        node {
                            name
                        }
                    }
                }
            }
        """

    result = schema.execute_sync(query)
    assert result.data == {"users": {"edges": [{"node": {"name": "John"}}]}}
    assert not result.errors


def test_lazy_optional_connection():
    @strawberry.type
    class Query:
        @strawberry.relay.connection(
            Optional[
                Annotated[
                    "EmptyUserConnection",
                    strawberry.lazy("tests.relay.test_connection"),
                ]
            ]
        )
        def users(self) -> list[User] | None:
            return None

    schema = strawberry.Schema(query=Query)
    query = """
            query {
                users {
                    edges {
                        node {
                            name
                        }
                    }
                }
            }
        """

    result = schema.execute_sync(query)
    assert result.data == {"users": None}
    assert not result.errors


def test_nullable_connection_with_pipe():
    @strawberry.type
    class Query:
        @strawberry.relay.connection(EmptyUserConnection | None)
        def users(self) -> list[User] | None:
            return None

    schema = strawberry.Schema(query=Query)
    query = """
        query {
            users {
                edges {
                    node {
                        name
                    }
                }
            }
        }
    """

    result = schema.execute_sync(query)
    assert result.data == {"users": None}
    assert not result.errors


def test_nullable_connection_with_permission():
    @strawberry.type
    class Query:
        @strawberry.relay.connection(
            Optional[EmptyUserConnection], permission_classes=[TestPermission]
        )
        def users(self) -> list[User] | None:  # pragma: no cover
            pytest.fail("Should not have been called...")

    schema = strawberry.Schema(query=Query)
    query = """
        query {
            users {
                edges {
                    node {
                        name
                    }
                }
            }
        }
    """

    result = schema.execute_sync(query)
    assert result.data == {"users": None}
    assert result.errors[0].message == "Not allowed"


@pytest.mark.parametrize(
    ("field_max_results", "schema_max_results", "results", "expected"),
    [
        (5, 100, 5, 5),
        (5, 2, 5, 5),
        (5, 100, 10, 5),
        (5, 2, 10, 5),
        (5, 100, 0, 0),
        (5, 2, 0, 0),
        (None, 100, 5, 5),
        (None, 2, 5, 2),
    ],
)
def test_max_results(
    field_max_results: int | None,
    schema_max_results: int,
    results: int,
    expected: int,
):
    @strawberry.type
    class User(Node):
        id: strawberry.relay.NodeID[str]

    @strawberry.type
    class Query:
        @strawberry.relay.connection(
            ListConnection[User],
            max_results=field_max_results,
        )
        def users(self) -> list[User]:
            return [User(id=str(i)) for i in range(results)]

    schema = strawberry.Schema(
        query=Query,
        config=StrawberryConfig(relay_max_results=schema_max_results),
    )
    query = """
      query {
        users {
          edges {
            node {
              id
            }
          }
        }
      }
    """

    result = schema.execute_sync(query)
    assert result.data is not None
    assert isinstance(result.data["users"]["edges"], list)
    assert len(result.data["users"]["edges"]) == expected



from __future__ import annotations

import concurrent.futures
from dataclasses import dataclass

from .mypy import run_mypy
from .pyright import run_pyright
from .result import Result
from .ty import run_ty


@dataclass
class TypecheckResult:
    pyright: list[Result]
    mypy: list[Result]
    ty: list[Result]


def typecheck(
    code: str, strict: bool = True, mypy_plugins: list[str] | None = None
) -> TypecheckResult:
    with concurrent.futures.ThreadPoolExecutor() as executor:
        pyright_future = executor.submit(run_pyright, code, strict=strict)
        mypy_future = executor.submit(
            run_mypy, code, strict=strict, mypy_plugins=mypy_plugins
        )
        ty_future = executor.submit(run_ty, code, strict=strict)

        pyright_results = pyright_future.result()
        mypy_results = mypy_future.result()
        ty_results = ty_future.result()

    return TypecheckResult(pyright=pyright_results, mypy=mypy_results, ty=ty_results)



# type: ignore

import textwrap

import strawberry
from strawberry.schema.config import StrawberryConfig
from strawberry.schema_directive import Location


def test_additional_schema_directives_printed_correctly_object():
    @strawberry.schema_directive(locations=[Location.OBJECT])
    class CacheControl:
        max_age: int

    @strawberry.federation.type(
        keys=["id"], shareable=True, extend=True, directives=[CacheControl(max_age=42)]
    )
    class FederatedType:
        id: strawberry.ID

    @strawberry.type
    class Query:
        federatedType: FederatedType  # noqa: N815

    expected_type = """
    directive @CacheControl(max_age: Int!) on OBJECT

    schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@key", "@shareable"]) {
      query: Query
    }

    extend type FederatedType @CacheControl(max_age: 42) @key(fields: "id") @shareable {
      id: ID!
    }

    type Query {
      _entities(representations: [_Any!]!): [_Entity]!
      _service: _Service!
      federatedType: FederatedType!
    }

    scalar _Any

    union _Entity = FederatedType

    type _Service {
      sdl: String!
    }
    """

    schema = strawberry.federation.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )
    assert schema.as_str() == textwrap.dedent(expected_type).strip()


def test_additional_schema_directives_printed_in_order_object():
    @strawberry.schema_directive(locations=[Location.OBJECT])
    class CacheControl0:
        max_age: int

    @strawberry.schema_directive(locations=[Location.OBJECT])
    class CacheControl1:
        min_age: int

    @strawberry.federation.type(
        keys=["id"],
        shareable=True,
        extend=True,
        directives=[CacheControl0(max_age=42), CacheControl1(min_age=42)],
    )
    class FederatedType:
        id: strawberry.ID

    @strawberry.type
    class Query:
        federatedType: FederatedType  # noqa: N815

    expected_type = """
    directive @CacheControl0(max_age: Int!) on OBJECT

    directive @CacheControl1(min_age: Int!) on OBJECT

    schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@key", "@shareable"]) {
      query: Query
    }

    extend type FederatedType @CacheControl0(max_age: 42) @CacheControl1(min_age: 42) @key(fields: "id") @shareable {
      id: ID!
    }

    type Query {
      _entities(representations: [_Any!]!): [_Entity]!
      _service: _Service!
      federatedType: FederatedType!
    }

    scalar _Any

    union _Entity = FederatedType

    type _Service {
      sdl: String!
    }
    """

    schema = strawberry.federation.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )
    assert schema.as_str() == textwrap.dedent(expected_type).strip()



import textwrap
from enum import Enum
from typing import Generic, TypeVar

import pytest

import strawberry
from strawberry.exceptions import DuplicatedTypeName
from strawberry.schema.config import StrawberryConfig


@pytest.mark.raises_strawberry_exception(
    DuplicatedTypeName,
    match=r"Type (.*) is defined multiple times in the schema",
)
def test_schema_has_no_duplicated_input_types():
    @strawberry.input(name="DuplicatedInput")
    class A:
        a: int

    @strawberry.input(name="DuplicatedInput")
    class B:
        b: int

    @strawberry.type
    class Query:
        field: int

    strawberry.Schema(query=Query, types=[A, B])


@pytest.mark.raises_strawberry_exception(
    DuplicatedTypeName,
    match=r"Type (.*) is defined multiple times in the schema",
)
def test_schema_has_no_duplicated_types():
    @strawberry.type(name="DuplicatedType")
    class A:
        a: int

    @strawberry.type(name="DuplicatedType")
    class B:
        b: int

    @strawberry.type
    class Query:
        field: int

    strawberry.Schema(query=Query, types=[A, B])


@pytest.mark.raises_strawberry_exception(
    DuplicatedTypeName,
    match=r"Type (.*) is defined multiple times in the schema",
)
def test_schema_has_no_duplicated_interfaces():
    @strawberry.interface(name="DuplicatedType")
    class A:
        a: int

    @strawberry.interface(name="DuplicatedType")
    class B:
        b: int

    @strawberry.type
    class Query:
        pass

    strawberry.Schema(query=Query, types=[A, B])


@pytest.mark.raises_strawberry_exception(
    DuplicatedTypeName,
    match=r"Type (.*) is defined multiple times in the schema",
)
def test_schema_has_no_duplicated_enums():
    @strawberry.enum(name="DuplicatedType")
    class A(Enum):
        A = 1

    @strawberry.enum(name="DuplicatedType")
    class B(Enum):
        B = 1

    @strawberry.type
    class Query:
        field: int

    strawberry.Schema(query=Query, types=[A, B])


@pytest.mark.raises_strawberry_exception(
    DuplicatedTypeName,
    match=r"Type (.*) is defined multiple times in the schema",
)
def test_schema_has_no_duplicated_names_across_different_types():
    @strawberry.interface(name="DuplicatedType")
    class A:
        a: int

    @strawberry.type(name="DuplicatedType")
    class B:
        b: int

    @strawberry.type
    class Query:
        field: int

    strawberry.Schema(query=Query, types=[A, B])


@pytest.mark.raises_strawberry_exception(
    DuplicatedTypeName,
    match=r"Type (.*) is defined multiple times in the schema",
)
def test_schema_has_no_duplicated_types_between_schema_and_extra_types():
    @strawberry.type(name="DuplicatedType")
    class A:
        a: int

    @strawberry.type(name="DuplicatedType")
    class B:
        b: int

    @strawberry.type
    class Query:
        field: A

    strawberry.Schema(query=Query, types=[B])


def test_allows_multiple_instance_of_same_generic():
    T = TypeVar("T")

    @strawberry.type
    class A(Generic[T]):
        a: T

    @strawberry.type
    class Query:
        first: A[int]
        second: A[int]

    schema = strawberry.Schema(Query)

    expected_schema = textwrap.dedent(
        """
        type IntA {
          a: Int!
        }

        type Query {
          first: IntA!
          second: IntA!
        }
        """
    ).strip()

    assert str(schema) == expected_schema


def test_allows_multiple_levels_of_nested_generics():
    T = TypeVar("T")

    @strawberry.type
    class Collection(Generic[T]):
        field1: list[T] = strawberry.field(resolver=lambda: [])  # noqa: PIE807
        field2: list[T] = strawberry.field(resolver=lambda: [])  # noqa: PIE807

    @strawberry.type
    class Container(Generic[T]):
        items: list[T]

    @strawberry.type
    class TypeA:
        id: str

    @strawberry.type
    class TypeB:
        id: int

    @strawberry.type
    class Query:
        @strawberry.field
        def a(self) -> Container[Collection[TypeA]]:
            return Container(items=[])

        @strawberry.field
        def b(self) -> Container[Collection[TypeB]]:
            return Container(items=[])

    schema = strawberry.Schema(query=Query)

    expected_schema = textwrap.dedent(
        """
        type Query {
          a: TypeACollectionContainer!
          b: TypeBCollectionContainer!
        }

        type TypeA {
          id: String!
        }

        type TypeACollection {
          field1: [TypeA!]!
          field2: [TypeA!]!
        }

        type TypeACollectionContainer {
          items: [TypeACollection!]!
        }

        type TypeB {
          id: Int!
        }

        type TypeBCollection {
          field1: [TypeB!]!
          field2: [TypeB!]!
        }

        type TypeBCollectionContainer {
          items: [TypeBCollection!]!
        }
        """
    ).strip()

    assert str(schema) == expected_schema


def test_allows_duplicated_types_when_validation_disabled():
    @strawberry.type(name="DuplicatedType")
    class A:
        a: int

    @strawberry.type(name="DuplicatedType")
    class B:
        b: int

    @strawberry.type
    class Query:
        field: int

    schema = strawberry.Schema(
        query=Query,
        types=[A, B],
        config=StrawberryConfig(_unsafe_disable_same_type_validation=True),
    )

    expected_schema = textwrap.dedent(
        """
        type DuplicatedType {
          a: Int!
        }

        type Query {
          field: Int!
        }
        """
    ).strip()

    assert str(schema) == expected_schema



"""
This file essentially mirrors the GraphQL over HTTP audits:
https://github.com/graphql/graphql-http/blob/main/src/audits/server.ts
"""

import pytest

try:
    from tests.http.clients.chalice import ChaliceHttpClient
except ImportError:
    ChaliceHttpClient = type(None)

try:
    from tests.http.clients.django import DjangoHttpClient
except ImportError:
    DjangoHttpClient = type(None)

try:
    from tests.http.clients.sanic import SanicHttpClient
except ImportError:
    SanicHttpClient = type(None)


@pytest.mark.xfail(
    reason="Our integrations currently only return application/json",
    raises=AssertionError,
)
async def test_22eb(http_client):
    """
    SHOULD accept application/graphql-response+json and match the content-type
    """
    response = await http_client.query(
        method="post",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        query="{ __typename }",
    )
    assert response.status_code == 200
    assert "application/graphql-response+json" in response.headers["content-type"]


async def test_4655(http_client):
    """
    MUST accept application/json and match the content-type
    """
    response = await http_client.query(
        method="post",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        query="{ __typename }",
    )
    assert response.status_code == 200
    assert "application/json" in response.headers["content-type"]


async def test_47de(http_client):
    """
    SHOULD accept */* and use application/json for the content-type
    """
    response = await http_client.query(
        method="post",
        headers={
            "Content-Type": "application/json",
            "Accept": "*/*",
        },
        query="{ __typename }",
    )
    assert response.status_code == 200
    assert "application/json" in response.headers["content-type"]


async def test_80d8(http_client):
    """
    SHOULD assume application/json content-type when accept is missing
    """
    response = await http_client.query(
        method="post",
        headers={"Content-Type": "application/json"},
        query="{ __typename }",
    )
    assert response.status_code == 200
    assert "application/json" in response.headers["content-type"]


async def test_82a3(http_client):
    """
    MUST use utf-8 encoding when responding
    """
    response = await http_client.query(
        method="post",
        headers={"Content-Type": "application/json"},
        query="{ __typename }",
    )
    assert response.status_code == 200
    assert isinstance(response.data, bytes)

    try:
        response.data.decode(encoding="utf-8", errors="strict")
    except UnicodeDecodeError:
        pytest.fail("Response body is not UTF-8 encoded")


async def test_bf61(http_client):
    """
    MUST accept utf-8 encoded request
    """
    response = await http_client.query(
        method="post",
        headers={"Content-Type": "application/json; charset=utf-8"},
        query='{ __type(name: "Run🏃Swim🏊") { name } }',
    )
    assert response.status_code == 200


async def test_78d5(http_client):
    """
    MUST assume utf-8 in request if encoding is unspecified
    """
    response = await http_client.query(
        method="post",
        headers={"Content-Type": "application/json"},
        query="{ __typename }",
    )
    assert response.status_code == 200


async def test_2c94(http_client):
    """
    MUST accept POST requests
    """
    response = await http_client.query(
        method="post",
        headers={"Content-Type": "application/json"},
        query="{ __typename }",
    )
    assert response.status_code == 200


async def test_5a70(http_client):
    """
    MAY accept application/x-www-form-urlencoded formatted GET requests
    """
    response = await http_client.query(method="get", query="{ __typename }")
    assert response.status_code == 200


async def test_9c48(http_client):
    """
    MAY NOT allow executing mutations on GET requests
    """
    response = await http_client.query(
        method="get",
        headers={"Accept": "application/graphql-response+json"},
        query="mutation { __typename }",
    )
    assert 400 <= response.status_code <= 499


@pytest.mark.xfail(
    reason="OPTIONAL - currently supported by Channels, Chalice, Django, and Sanic",
    raises=AssertionError,
)
async def test_9abe(http_client):
    """
    MAY respond with 4xx status code if content-type is not supplied on POST requests
    """
    response = await http_client.post(
        url="/graphql",
        headers={},
        json={"query": "{ __typename }"},
    )
    assert 400 <= response.status_code <= 499


async def test_03d4(http_client):
    """
    MUST accept application/json POST requests
    """
    response = await http_client.query(
        method="post",
        headers={"Content-Type": "application/json"},
        query="{ __typename }",
    )
    assert response.status_code == 200


async def test_a5bf(http_client):
    """
    MAY use 400 status code when request body is missing on POST
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
    )
    assert response.status_code == 400


async def test_423l(http_client):
    """
    MAY use 400 status code on missing {query} parameter
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={"notquery": "{ __typename }"},
    )
    assert response.status_code == 400


@pytest.mark.parametrize(
    "invalid",
    [{"obj": "ect"}, 0, False, ["array"]],
    ids=["LKJ0", "LKJ1", "LKJ2", "LKJ3"],
)
async def test_lkj_(http_client, invalid):
    """
    MAY use 400 status code on invalid {query} parameter
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
        json={"query": invalid},
    )
    assert response.status_code == 400


async def test_34a2(http_client):
    """
    SHOULD allow string {query} parameter when accepting application/graphql-response+json
    """
    response = await http_client.query(
        method="post",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        query="{ __typename }",
    )
    assert response.status_code == 200


async def test_13ee(http_client):
    """
    MUST allow string {query} parameter when accepting application/json
    """
    response = await http_client.query(
        method="post",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        query="{ __typename }",
    )
    assert response.status_code == 200
    assert isinstance(response.json, dict)
    assert "errors" not in response.json


@pytest.mark.parametrize(
    "invalid",
    [{"obj": "ect"}, 0, False, ["array"]],
    ids=["6C00", "6C01", "6C02", "6C03"],
)
async def test_6c0_(http_client, invalid):
    """
    MAY use 400 status code on invalid {operationName} parameter
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
        json={
            "operationName": invalid,
            "query": "{ __typename }",
        },
    )
    assert response.status_code == 400


async def test_8161(http_client):
    """
    SHOULD allow string {operationName} parameter when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={
            "operationName": "Query",
            "query": "query Query { __typename }",
        },
    )
    assert response.status_code == 200


async def test_b8b3(http_client):
    """
    MUST allow string {operationName} parameter when accepting application/json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "operationName": "Query",
            "query": "query Query { __typename }",
        },
    )
    assert response.status_code == 200
    assert isinstance(response.json, dict)
    assert "errors" not in response.json


@pytest.mark.parametrize(
    "parameter",
    ["variables", "operationName", "extensions"],
    ids=["94B0", "94B1", "94B2"],
)
async def test_94b_(http_client, parameter):
    """
    SHOULD allow null variables/operationName/extensions parameter when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={
            "query": "{ __typename }",
            parameter: None,
        },
    )
    assert response.status_code == 200
    assert "errors" not in response.json


@pytest.mark.parametrize(
    "parameter",
    ["variables", "operationName", "extensions"],
    ids=["0220", "0221", "0222"],
)
async def test_022_(http_client, parameter):
    """
    MUST allow null variables/operationName/extensions parameter when accepting application/json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "query": "{ __typename }",
            parameter: None,
        },
    )
    assert response.status_code == 200
    assert "errors" not in response.json


@pytest.mark.parametrize(
    "invalid",
    ["string", 0, False, ["array"]],
    ids=["4760", "4761", "4762", "4763"],
)
async def test_476_(http_client, invalid):
    """
    MAY use 400 status code on invalid {variables} parameter
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
        json={
            "query": "{ __typename }",
            "variables": invalid,
        },
    )
    assert response.status_code == 400


async def test_2ea1(http_client):
    """
    SHOULD allow map {variables} parameter when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={
            "query": "query Type($name: String!) { __type(name: $name) { name } }",
            "variables": {"name": "sometype"},
        },
    )
    assert response.status_code == 200


async def test_28b9(http_client):
    """
    MUST allow map {variables} parameter when accepting application/json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "query": "query Type($name: String!) { __type(name: $name) { name } }",
            "variables": {"name": "sometype"},
        },
    )
    assert response.status_code == 200
    assert isinstance(response.json, dict)
    assert "errors" not in response.json


async def test_d6d5(http_client):
    """
    MAY allow URL-encoded JSON string {variables} parameter in GETs when accepting application/graphql-response+json
    """
    response = await http_client.query(
        query="query Type($name: String!) { __type(name: $name) { name } }",
        variables={"name": "sometype"},
        method="get",
        headers={"Accept": "application/graphql-response+json"},
    )
    assert response.status_code == 200


async def test_6a70(http_client):
    """
    MAY allow URL-encoded JSON string {variables} parameter in GETs when accepting application/json
    """
    response = await http_client.query(
        query="query Type($name: String!) { __type(name: $name) { name } }",
        variables={"name": "sometype"},
        method="get",
        headers={"Accept": "application/json"},
    )
    assert response.status_code == 200
    assert isinstance(response.json, dict)
    assert "errors" not in response.json


@pytest.mark.parametrize(
    "invalid",
    ["string", 0, False, ["array"]],
    ids=["58B0", "58B1", "58B2", "58B3"],
)
async def test_58b_(http_client, invalid):
    """
    MAY use 400 status code on invalid {extensions} parameter
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
        json={
            "query": "{ __typename }",
            "extensions": invalid,
        },
    )
    assert response.status_code == 400


async def test_428f(http_client):
    """
    SHOULD allow map {extensions} parameter when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={
            "query": "{ __typename }",
            "extensions": {"some": "value"},
        },
    )
    assert response.status_code == 200


async def test_1b7a(http_client):
    """
    MUST allow map {extensions} parameter when accepting application/json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "query": "{ __typename }",
            "extensions": {"some": "value"},
        },
    )
    assert response.status_code == 200
    assert isinstance(response.json, dict)
    assert "errors" not in response.json


async def test_b6dc(http_client):
    """
    MAY use 4xx or 5xx status codes on JSON parsing failure
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
        data=b'{ "not a JSON',
    )
    assert 400 <= response.status_code <= 599


async def test_bcf8(http_client):
    """
    MAY use 400 status code on JSON parsing failure
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
        data=b'{ "not a JSON',
    )
    assert response.status_code == 400


async def test_8764(http_client):
    """
    MAY use 4xx or 5xx status codes if parameters are invalid
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
        json={"qeury": "{ __typename }"},  # typo in 'query'
    )
    assert 400 <= response.status_code <= 599


async def test_3e3a(http_client):
    """
    MAY use 400 status code if parameters are invalid
    """
    response = await http_client.post(
        url="/graphql",
        headers={"Content-Type": "application/json"},
        json={"qeury": "{ __typename }"},  # typo in 'query'
    )
    assert response.status_code == 400


async def test_39aa(http_client):
    """
    MUST accept a map for the {extensions} parameter
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "query": "{ __typename }",
            "extensions": {"some": "value"},
        },
    )
    assert response.status_code == 200
    assert isinstance(response.json, dict)
    assert "errors" not in response.json


async def test_572b(http_client):
    """
    SHOULD use 200 status code on document parsing failure when accepting application/json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={"query": "{"},
    )
    assert response.status_code == 200


async def test_dfe2(http_client):
    """
    SHOULD use 200 status code on document validation failure when accepting application/json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "query": "{ 8f31403dfe404bccbb0e835f2629c6a7 }"
        },  # making sure the field doesn't exist
    )
    assert response.status_code == 200


async def test_7b9b(http_client):
    """
    SHOULD use a status code of 200 on variable coercion failure when accepting application/json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "query": "query CoerceFailure($id: ID!){ __typename }",
            "variables": {"id": None},
        },
    )
    assert response.status_code == 200


@pytest.mark.xfail(
    reason="Currently results in status 200 with GraphQL errors", raises=AssertionError
)
async def test_865d(http_client):
    """
    SHOULD use 4xx or 5xx status codes on document parsing failure when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={"query": "{"},
    )
    assert 400 <= response.status_code <= 599


@pytest.mark.xfail(
    reason="Currently results in status 200 with GraphQL errors", raises=AssertionError
)
async def test_556a(http_client):
    """
    SHOULD use 400 status code on document parsing failure when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={"query": "{"},
    )
    assert response.status_code == 400


@pytest.mark.xfail(
    reason="Currently results in status 200 with GraphQL errors", raises=AssertionError
)
async def test_d586(http_client):
    """
    SHOULD NOT contain the data entry on document parsing failure when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={"query": "{"},
    )
    assert response.status_code == 400
    assert "data" not in response.json


@pytest.mark.xfail(
    reason="Currently results in status 200 with GraphQL errors", raises=AssertionError
)
async def test_51fe(http_client):
    """
    SHOULD use 4xx or 5xx status codes on document validation failure when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={
            "query": "{ 8f31403dfe404bccbb0e835f2629c6a7 }",  # making sure the field doesn't exist
        },
    )
    assert 400 <= response.status_code <= 599


@pytest.mark.xfail(
    reason="Currently results in status 200 with GraphQL errors", raises=AssertionError
)
async def test_74ff(http_client):
    """
    SHOULD use 400 status code on document validation failure when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={
            "query": "{ 8f31403dfe404bccbb0e835f2629c6a7 }",  # making sure the field doesn't exist
        },
    )
    assert response.status_code == 400


@pytest.mark.xfail(
    reason="Currently results in status 200 with GraphQL errors", raises=AssertionError
)
async def test_5e5b(http_client):
    """
    SHOULD NOT contain the data entry on document validation failure when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={
            "query": "{ 8f31403dfe404bccbb0e835f2629c6a7 }",  # making sure the field doesn't exist
        },
    )
    assert response.status_code == 400
    assert "data" not in response.json


@pytest.mark.xfail(
    reason="Currently results in status 200 with GraphQL errors", raises=AssertionError
)
async def test_86ee(http_client):
    """
    SHOULD use a status code of 400 on variable coercion failure when accepting application/graphql-response+json
    """
    response = await http_client.post(
        url="/graphql",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/graphql-response+json",
        },
        json={
            "query": "query CoerceFailure($id: ID!){ __typename }",
            "variables": {"id": None},
        },
    )
    assert response.status_code == 400



import dataclasses
import json
from typing import Annotated, Optional
from unittest.mock import MagicMock

import pytest

import strawberry
from strawberry.types.base import StrawberryOptional
from strawberry.types.info import Info
from strawberry.types.nodes import FragmentSpread, InlineFragment, SelectedField
from strawberry.types.unset import UNSET


def test_info_has_the_correct_shape():
    my_context = "123"
    root_value = "ABC"

    @strawberry.type
    class Result:
        field_name: str
        python_name: str
        selected_field: str
        operation: str
        path: str
        variable_values: str
        context_equal: bool
        root_equal: bool
        return_type: str
        schema_print: str

    @strawberry.type
    class Query:
        @strawberry.field
        def hello_world(self, info: strawberry.Info[str, str]) -> Result:
            return Result(
                path="".join([str(p) for p in info.path.as_list()]),
                operation=str(info.operation),
                field_name=info.field_name,
                python_name=info.python_name,
                selected_field=json.dumps(dataclasses.asdict(*info.selected_fields)),
                variable_values=str(info.variable_values),
                context_equal=info.context == my_context,
                root_equal=info.root_value == root_value,
                return_type=str(info.return_type),
                schema_print=info.schema.as_str(),
            )

    schema = strawberry.Schema(query=Query)

    query = """{
        helloWorld {
            fieldName
            pythonName
            selectedField
            contextEqual
            operation
            path
            rootEqual
            variableValues
            returnType
            schemaPrint
        }
    }"""

    result = schema.execute_sync(query, context_value=my_context, root_value=root_value)

    assert not result.errors
    assert result.data
    info = result.data["helloWorld"]
    assert info.pop("operation").startswith("OperationDefinitionNode at")
    field = json.loads(info.pop("selectedField"))
    selections = {selection["name"] for selection in field.pop("selections")}
    assert selections == {
        "selectedField",
        "path",
        "rootEqual",
        "operation",
        "contextEqual",
        "variableValues",
        "returnType",
        "fieldName",
        "pythonName",
        "schemaPrint",
    }
    assert field == {
        "name": "helloWorld",
        "directives": {},
        "alias": None,
        "arguments": {},
    }
    assert info == {
        "fieldName": "helloWorld",
        "pythonName": "hello_world",
        "path": "helloWorld",
        "contextEqual": True,
        "rootEqual": True,
        "variableValues": "{}",
        "returnType": "<class 'tests.schema.test_info.test_info_has_the_correct_shape.<locals>.Result'>",
        "schemaPrint": schema.as_str(),
    }


def test_info_field_fragments():
    @strawberry.type
    class Result:
        ok: bool

    selected_fields = None

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, info: strawberry.Info[str, str]) -> Result:
            nonlocal selected_fields
            selected_fields = info.selected_fields
            return Result(ok=True)

    schema = strawberry.Schema(query=Query)
    query = """{
        hello {
            ... on Result {
                k: ok @include(if: true)
            }
            ...frag
        }
    }

    fragment frag on Result {
        ok
    }
    """
    result = schema.execute_sync(query)

    assert not result.errors
    assert selected_fields == [
        SelectedField(
            name="hello",
            directives={},
            alias=None,
            arguments={},
            selections=[
                InlineFragment(
                    type_condition="Result",
                    directives={},
                    selections=[
                        SelectedField(
                            name="ok",
                            alias="k",
                            arguments={},
                            directives={
                                "include": {
                                    "if": True,
                                },
                            },
                            selections=[],
                        )
                    ],
                ),
                FragmentSpread(
                    name="frag",
                    directives={},
                    type_condition="Result",
                    selections=[
                        SelectedField(
                            name="ok",
                            directives={},
                            arguments={},
                            selections=[],
                        )
                    ],
                ),
            ],
        )
    ]


def test_info_arguments():
    @strawberry.input
    class TestInput:
        name: str
        age: int | None = UNSET

    selected_fields = None

    @strawberry.type
    class Query:
        @strawberry.field
        def test_arg(
            self,
            info: strawberry.Info[str, str],
            input: TestInput,
            another_arg: bool = True,
        ) -> str:
            nonlocal selected_fields
            selected_fields = info.selected_fields
            return "Hi"

    schema = strawberry.Schema(query=Query)

    query = """{
        testArg(input: {name: "hi"})
    }
    """
    result = schema.execute_sync(query)

    assert not result.errors
    assert selected_fields == [
        SelectedField(
            name="testArg",
            directives={},
            arguments={
                "input": {
                    "name": "hi",
                },
            },
            selections=[],
        )
    ]

    query = """query TestQuery($input: TestInput!) {
        testArg(input: $input)
    }
    """
    result = schema.execute_sync(
        query,
        variable_values={
            "input": {
                "name": "hi",
                "age": 10,
            },
        },
    )
    assert not result.errors
    assert selected_fields == [
        SelectedField(
            name="testArg",
            directives={},
            arguments={
                "input": {
                    "name": "hi",
                    "age": 10,
                },
            },
            selections=[],
        )
    ]


def test_info_selected_fields_undefined_variable():
    @strawberry.type
    class Result:
        ok: bool

    selected_fields = None

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(
            self, info: strawberry.Info[str, str], optional_input: str | None = "hi"
        ) -> Result:
            nonlocal selected_fields
            selected_fields = info.selected_fields
            return Result(ok=True)

    schema = strawberry.Schema(query=Query)
    query = """
    query MyQuery($optionalInput: String) {
        hello(optionalInput: $optionalInput) {
            ok
        }
    }
    """
    result = schema.execute_sync(query, variable_values={})

    assert not result.errors
    assert selected_fields == [
        SelectedField(
            name="hello",
            directives={},
            alias=None,
            arguments={
                "optionalInput": None,
            },
            selections=[
                SelectedField(
                    name="ok",
                    alias=None,
                    arguments={},
                    directives={},
                    selections=[],
                )
            ],
        )
    ]


@pytest.mark.parametrize(
    ("return_type", "return_value"),
    [
        (str, "text"),
        (list[str], ["text"]),
        (Optional[list[int]], None),
    ],
)
def test_return_type_from_resolver(return_type, return_value):
    @strawberry.type
    class Query:
        @strawberry.field
        def field(self, info: strawberry.Info) -> return_type:
            assert info.return_type == return_type
            return return_value

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync("{ field }")

    assert not result.errors
    assert result.data
    assert result.data["field"] == return_value


def test_return_type_from_field():
    def resolver(info: strawberry.Info):
        assert info.return_type is int
        return 0

    @strawberry.type
    class Query:
        field: int = strawberry.field(resolver=resolver)

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync("{ field }")

    assert not result.errors
    assert result.data
    assert result.data["field"] == 0


def test_field_nodes_deprecation():
    def resolver(info: strawberry.Info):
        info.field_nodes
        return 0

    @strawberry.type
    class Query:
        field: int = strawberry.field(resolver=resolver)

    schema = strawberry.Schema(query=Query)

    with pytest.deprecated_call():
        result = schema.execute_sync("{ field }")

    assert not result.errors
    assert result.data
    assert result.data["field"] == 0


def test_info_query():
    graphql_query = None

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, info: strawberry.Info) -> str:
            nonlocal graphql_query
            graphql_query = info.query
            return "world"

    schema = strawberry.Schema(query=Query)
    query = "{ hello }"
    result = schema.execute_sync(query)

    assert not result.errors
    assert graphql_query == query


def test_info_query_with_variables():
    graphql_query = None

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, info: strawberry.Info, name: str = "world") -> str:
            nonlocal graphql_query
            graphql_query = info.query
            return f"hello {name}"

    schema = strawberry.Schema(query=Query)
    query = """query Hello($name: String!) {
        hello(name: $name)
    }"""
    result = schema.execute_sync(query, variable_values={"name": "test"})

    assert not result.errors
    assert graphql_query == query


def test_info_query_with_fragments():
    graphql_query = None

    @strawberry.type
    class User:
        name: str
        age: int

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self, info: strawberry.Info) -> User:
            nonlocal graphql_query
            graphql_query = info.query
            return User(name="Alice", age=30)

    schema = strawberry.Schema(query=Query)
    query = """query {
        user {
            ...UserFields
        }
    }

    fragment UserFields on User {
        name
        age
    }"""
    result = schema.execute_sync(query)

    assert not result.errors
    assert graphql_query == query


def test_info_query_with_mutation():
    graphql_query = None

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def create_user(self, info: strawberry.Info, name: str) -> str:
            nonlocal graphql_query
            graphql_query = info.query
            return name

    schema = strawberry.Schema(query=Query, mutation=Mutation)
    query = """mutation CreateUser($name: String!) {
        createUser(name: $name)
    }"""
    result = schema.execute_sync(query, variable_values={"name": "Alice"})

    assert not result.errors
    assert graphql_query == query


def test_info_query_with_multiple_operations():
    graphql_query = None

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, info: strawberry.Info) -> str:
            nonlocal graphql_query
            graphql_query = info.query
            return "world"

        @strawberry.field
        def goodbye(self, info: strawberry.Info) -> str:
            nonlocal graphql_query
            graphql_query = info.query
            return "farewell"

    schema = strawberry.Schema(query=Query)
    query = """query FirstOp {
        hello
    }

    query SecondOp {
        goodbye
    }"""
    result = schema.execute_sync(query, operation_name="SecondOp")

    assert not result.errors
    # info.query returns the full document, including all operations
    assert graphql_query == query
    assert "FirstOp" in graphql_query
    assert "SecondOp" in graphql_query


def test_info_query_returns_none_without_loc():
    mock_raw_info = MagicMock()
    mock_raw_info.operation.loc = None

    mock_field = MagicMock()

    info = Info(_raw_info=mock_raw_info, _field=mock_field)
    assert info.query is None


def test_get_argument_defintion_helper():
    @strawberry.input
    class TestInput:
        foo: str

    arg_1_def = None
    arg_2_def = None
    missing_arg_def = None

    @strawberry.type
    class Query:
        @strawberry.field
        def field(
            self,
            info: strawberry.Info,
            arg_1: Annotated[str, strawberry.argument(description="Some description")],
            arg_2: TestInput | None = None,
        ) -> str:
            nonlocal arg_1_def, arg_2_def, missing_arg_def
            arg_1_def = info.get_argument_definition("arg_1")
            arg_2_def = info.get_argument_definition("arg_2")
            missing_arg_def = info.get_argument_definition("missing_arg_def")

            return "bar"

    schema = strawberry.Schema(query=Query)
    result = schema.execute_sync('{ field(arg1: "hi") }')

    assert not result.errors
    assert arg_1_def
    assert arg_1_def.type is str
    assert arg_1_def.python_name == "arg_1"
    assert arg_1_def.description == "Some description"

    assert arg_2_def
    assert arg_2_def.python_name == "arg_2"
    assert arg_2_def.default is None
    assert isinstance(arg_2_def.type, StrawberryOptional)
    assert arg_2_def.type.of_type is TestInput

    assert missing_arg_def is None



import strawberry

# create a few types and then a union type


@strawberry.type
class Foo:
    a: str


@strawberry.type
class Bar:
    b: str


@strawberry.type
class Baz:
    c: str


@strawberry.type
class Qux:
    d: str


# this is the union type

Union1 = strawberry.union(name="Union1", types=(Foo, Bar, Baz, Qux))
Union2 = strawberry.union(name="Union2", types=(Baz, Qux))



from unittest.mock import patch

import pytest
from graphql import SourceLocation, parse

import strawberry
from strawberry.extensions import MaxTokensLimiter, ParserCache


@patch("strawberry.extensions.parser_cache.parse", wraps=parse)
def test_parser_cache_extension(mock_parse):
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

        @strawberry.field
        def ping(self) -> str:
            return "pong"

    schema = strawberry.Schema(query=Query, extensions=[ParserCache()])

    query = "query { hello }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"hello": "world"}

    assert mock_parse.call_count == 1

    # Run query multiple times
    for _ in range(3):
        result = schema.execute_sync(query)
        assert not result.errors
        assert result.data == {"hello": "world"}

    # validate is still only called once
    assert mock_parse.call_count == 1

    # Running a second query doesn't cache
    query2 = "query { ping }"
    result = schema.execute_sync(query2)

    assert not result.errors
    assert result.data == {"ping": "pong"}

    assert mock_parse.call_count == 2


@patch("strawberry.extensions.parser_cache.parse", wraps=parse)
def test_parser_cache_extension_arguments(mock_parse):
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

        @strawberry.field
        def ping(self) -> str:
            return "pong"

    schema = strawberry.Schema(
        query=Query, extensions=[MaxTokensLimiter(max_token_count=20), ParserCache()]
    )

    query = "query { hello }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"hello": "world"}

    mock_parse.assert_called_with("query { hello }", max_tokens=20)


@patch("strawberry.extensions.parser_cache.parse", wraps=parse)
def test_parser_cache_extension_syntax_error(mock_parse):
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:  # pragma: no cover
            return "world"

    schema = strawberry.Schema(query=Query, extensions=[ParserCache()])

    query = "query { hello"

    result = schema.execute_sync(query)

    assert len(result.errors) == 1
    assert result.errors[0].message == "Syntax Error: Expected Name, found <EOF>."
    assert result.errors[0].locations == [SourceLocation(line=1, column=14)]
    assert mock_parse.call_count == 1


@patch("strawberry.extensions.parser_cache.parse", wraps=parse)
def test_parser_cache_extension_max_size(mock_parse):
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

        @strawberry.field
        def ping(self) -> str:
            return "pong"

    schema = strawberry.Schema(query=Query, extensions=[ParserCache(maxsize=1)])

    query = "query { hello }"

    for _ in range(2):
        result = schema.execute_sync(query)
        assert not result.errors
        assert result.data == {"hello": "world"}

    assert mock_parse.call_count == 1

    query2 = "query { ping }"
    result = schema.execute_sync(query2)

    assert not result.errors
    assert mock_parse.call_count == 2

    # running the first query again doesn't cache
    result = schema.execute_sync(query)
    assert not result.errors

    # validate is still only called once
    assert mock_parse.call_count == 3


@pytest.mark.asyncio
@patch("strawberry.extensions.parser_cache.parse", wraps=parse)
async def test_parser_cache_extension_async(mock_parse):
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

        @strawberry.field
        def ping(self) -> str:
            return "pong"

    schema = strawberry.Schema(query=Query, extensions=[ParserCache()])

    query = "query { hello }"

    result = await schema.execute(query)

    assert not result.errors
    assert result.data == {"hello": "world"}

    assert mock_parse.call_count == 1

    # Run query multiple times
    for _ in range(3):
        result = await schema.execute(query)
        assert not result.errors
        assert result.data == {"hello": "world"}

    # validate is still only called once
    assert mock_parse.call_count == 1

    # Running a second query doesn't cache
    query2 = "query { ping }"
    result = await schema.execute(query2)

    assert not result.errors
    assert result.data == {"ping": "pong"}

    assert mock_parse.call_count == 2



import textwrap
from typing import Any
from typing_extensions import Self

import pytest

import strawberry
from strawberry import Schema, relay
from strawberry.relay import NodeType, to_base64


@pytest.fixture
def schema() -> Schema:
    @strawberry.type
    class Fruit(relay.Node):
        code: relay.NodeID[int]

    @strawberry.type(name="Edge", description="An edge in a connection.")
    class CustomEdge(relay.Edge[NodeType]):
        CURSOR_PREFIX = "customprefix"
        index: int

        @classmethod
        def resolve_edge(
            cls, node: NodeType, *, cursor: Any = None, **kwargs: Any
        ) -> Self:
            assert isinstance(cursor, int)
            return super().resolve_edge(node, cursor=cursor, index=cursor, **kwargs)

    @strawberry.type(name="Connection", description="A connection to a list of items.")
    class CustomListConnection(relay.ListConnection[NodeType]):
        edges: list[CustomEdge[NodeType]] = strawberry.field(
            description="Contains the nodes in this connection"
        )

    @strawberry.type
    class Query:
        @relay.connection(CustomListConnection[Fruit])
        def fruits(self) -> list[Fruit]:
            return [Fruit(code=i) for i in range(10)]

    return strawberry.Schema(query=Query)


def test_schema_with_custom_edge_class(schema: Schema):
    expected = textwrap.dedent(
        '''
        type Fruit implements Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        """A connection to a list of items."""
        type FruitConnection {
          """Pagination data for this connection"""
          pageInfo: PageInfo!

          """Contains the nodes in this connection"""
          edges: [FruitEdge!]!
        }

        """An edge in a connection."""
        type FruitEdge {
          """A cursor for use in pagination"""
          cursor: String!

          """The item at the end of the edge"""
          node: Fruit!
          index: Int!
        }

        """An object with a Globally Unique ID"""
        interface Node {
          """The Globally Unique ID of this object"""
          id: ID!
        }

        """Information to aid in pagination."""
        type PageInfo {
          """When paginating forwards, are there more items?"""
          hasNextPage: Boolean!

          """When paginating backwards, are there more items?"""
          hasPreviousPage: Boolean!

          """When paginating backwards, the cursor to continue."""
          startCursor: String

          """When paginating forwards, the cursor to continue."""
          endCursor: String
        }

        type Query {
          fruits(
            """Returns the items in the list that come before the specified cursor."""
            before: String = null

            """Returns the items in the list that come after the specified cursor."""
            after: String = null

            """Returns the first n items from the list."""
            first: Int = null

            """Returns the items in the list that come after the specified cursor."""
            last: Int = null
          ): FruitConnection!
        }
        '''
    ).strip()
    assert str(schema) == expected


def test_custom_cursor_prefix_used(schema: Schema):
    result = schema.execute_sync(
        """
        query {
          fruits {
            edges {
              cursor
              node {
                id
              }
            }
          }
        }
        """
    )
    assert result.errors is None
    assert result.data == {
        "fruits": {
            "edges": [
                {
                    "cursor": to_base64("customprefix", i),
                    "node": {"id": to_base64("Fruit", i)},
                }
                for i in range(10)
            ]
        }
    }


def test_custom_cursor_prefix_can_be_parsed(schema: Schema):
    result = schema.execute_sync(
        """
        query TestQuery($after: String) {
          fruits(after: $after) {
            edges {
              cursor
              node {
                id
              }
            }
          }
        }
        """,
        {"after": to_base64("customprefix", 2)},
    )
    assert result.errors is None
    assert result.data == {
        "fruits": {
            "edges": [
                {
                    "cursor": to_base64("customprefix", i),
                    "node": {"id": to_base64("Fruit", i)},
                }
                for i in range(3, 10)
            ]
        }
    }


def test_custom_edge_class_fields_can_be_resolved(schema: Schema):
    result = schema.execute_sync(
        """
        query TestQuery($after: String) {
          fruits(after: $after) {
            edges {
              index
              node {
                id
              }
            }
          }
        }
        """,
        {"after": to_base64("customprefix", 2)},
    )
    assert result.errors is None
    assert result.data == {
        "fruits": {
            "edges": [
                {"index": i, "node": {"id": to_base64("Fruit", i)}}
                for i in range(3, 10)
            ]
        }
    }



import base64
from typing import NewType

import strawberry

Base64Encoded = strawberry.scalar(
    NewType("Base64Encoded", bytes),
    serialize=base64.b64encode,
    parse_value=base64.b64decode,
)


@strawberry.scalar(serialize=lambda x: 42, parse_value=lambda x: Always42())
class Always42:
    pass


MyStr = strawberry.scalar(NewType("MyStr", str))


def test_custom_scalar_serialization():
    @strawberry.type
    class Query:
        @strawberry.field
        def custom_scalar_field(self) -> Base64Encoded:
            return Base64Encoded(b"decoded value")

    schema = strawberry.Schema(Query)

    result = schema.execute_sync("{ customScalarField }")

    assert not result.errors
    assert base64.b64decode(result.data["customScalarField"]) == b"decoded value"


def test_custom_scalar_deserialization():
    @strawberry.type
    class Query:
        @strawberry.field
        def decode_base64(self, encoded: Base64Encoded) -> str:
            return bytes(encoded).decode("ascii")

    schema = strawberry.Schema(Query)

    encoded = Base64Encoded(base64.b64encode(b"decoded"))
    query = """query decode($encoded: Base64Encoded!) {
        decodeBase64(encoded: $encoded)
    }"""
    result = schema.execute_sync(query, variable_values={"encoded": encoded})

    assert not result.errors
    assert result.data["decodeBase64"] == "decoded"


def test_custom_scalar_decorated_class():
    @strawberry.type
    class Query:
        @strawberry.field
        def answer(self) -> Always42:
            return Always42()

    schema = strawberry.Schema(Query)

    result = schema.execute_sync("{ answer }")

    assert not result.errors
    assert result.data["answer"] == 42


def test_custom_scalar_default_serialization():
    @strawberry.type
    class Query:
        @strawberry.field
        def my_str(self, arg: MyStr) -> MyStr:
            return MyStr(str(arg) + "Suffix")

    schema = strawberry.Schema(Query)

    result = schema.execute_sync('{ myStr(arg: "value") }')

    assert not result.errors
    assert result.data["myStr"] == "valueSuffix"



import strawberry


def test_basic_list():
    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> list[str]:
            return ["Example"]

    schema = strawberry.Schema(query=Query)

    query = "{ example }"

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["example"] == ["Example"]


def test_of_optional():
    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> list[str | None]:
            return ["Example", None]

    schema = strawberry.Schema(query=Query)

    query = "{ example }"

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["example"] == ["Example", None]


def test_lists_of_lists():
    def get_polygons() -> list[list[float]]:
        return [[2.0, 6.0]]

    @strawberry.type
    class Query:
        polygons: list[list[float]] = strawberry.field(resolver=get_polygons)

    schema = strawberry.Schema(query=Query)

    query = "{ polygons }"

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["polygons"] == [[2.0, 6.0]]



from enum import Enum

import strawberry
from strawberry import UNSET, Maybe, Some, asdict


def test_simple_type():
    @strawberry.type
    class User:
        name: str
        age: int

    assert asdict(User(name="Alex", age=30)) == {"name": "Alex", "age": 30}


def test_input_type():
    @strawberry.input
    class Input:
        title: str
        description: str
        tags: list[str] | None = strawberry.field(default=None)

    assert asdict(Input(title="foo", description="bar")) == {
        "title": "foo",
        "description": "bar",
        "tags": None,
    }


def test_type_with_nested_type_and_enum_in_list():
    @strawberry.enum
    class Count(Enum):
        TWO = "two"
        FOUR = "four"

    @strawberry.type
    class Animal:
        legs: Count

    @strawberry.type
    class People:
        name: str
        animals: list[Animal]

    assert asdict(
        People(name="Kevin", animals=[Animal(legs=Count.TWO), Animal(legs=Count.FOUR)])
    ) == {
        "name": "Kevin",
        "animals": [{"legs": Count.TWO}, {"legs": Count.FOUR}],
    }


def test_unset_field_is_excluded():
    @strawberry.input
    class Input:
        field: str | None

    assert asdict(Input(field=None)) == {"field": None}
    assert asdict(Input(field=UNSET)) == {}


def test_unset_field_in_nested_input_is_excluded():
    @strawberry.input
    class InnerInput:
        field: str | None

    @strawberry.input
    class OuterInput:
        nested: InnerInput

    assert asdict(OuterInput(nested=InnerInput(field=UNSET))) == {"nested": {}}


def test_maybe_field_absent_is_excluded():
    @strawberry.input
    class Input:
        field: Maybe[str | None]

    assert asdict(Input(field=None)) == {}


def test_maybe_field_unset_is_excluded():
    @strawberry.input
    class Input:
        field: Maybe[str | None]

    assert asdict(Input(field=UNSET)) == {}


def test_maybe_field_with_some_value():
    @strawberry.input
    class Input:
        field: Maybe[str | None]

    assert asdict(Input(field=Some("foo"))) == {"field": "foo"}


def test_maybe_field_with_some_none():
    @strawberry.input
    class Input:
        field: Maybe[str | None]

    assert asdict(Input(field=Some(None))) == {"field": None}


def test_multiple_maybe_fields_partial():
    @strawberry.input
    class Input:
        field_a: Maybe[str | None]
        field_b: Maybe[str | None]
        field_c: Maybe[str | None]

    # field_b is absent (None default), so it should be excluded
    assert asdict(
        Input(field_a=Some("hello"), field_b=None, field_c=Some("world"))
    ) == {
        "field_a": "hello",
        "field_c": "world",
    }


def test_nested_input():
    @strawberry.input
    class InnerInput:
        field: str

    @strawberry.input
    class OuterInput:
        scalar: str
        nested: InnerInput

    assert asdict(OuterInput(scalar="foo", nested=InnerInput(field="bar"))) == {
        "scalar": "foo",
        "nested": {"field": "bar"},
    }


def test_nested_input_three_levels_deep():
    @strawberry.input
    class Level3:
        field: str

    @strawberry.input
    class Level2:
        nested: Level3

    @strawberry.input
    class Level1:
        nested: Level2

    assert asdict(Level1(nested=Level2(nested=Level3(field="deep")))) == {
        "nested": {"nested": {"field": "deep"}}
    }


def test_optional_nested_input():
    @strawberry.input
    class InnerInput:
        field: str

    @strawberry.input
    class OuterInput:
        nested: InnerInput | None = None

    assert asdict(OuterInput()) == {"nested": None}
    assert asdict(OuterInput(nested=InnerInput(field="foo"))) == {
        "nested": {"field": "foo"}
    }


def test_maybe_nested_input():
    @strawberry.input
    class InnerInput:
        field: str

    @strawberry.input
    class OuterInput:
        nested: Maybe[InnerInput]

    assert asdict(OuterInput(nested=Some(InnerInput(field="foo")))) == {
        "nested": {"field": "foo"}
    }


def test_maybe_nested_input_explicitly_null():
    @strawberry.input
    class InnerInput:
        field: str

    @strawberry.input
    class OuterInput:
        nested: Maybe[InnerInput | None]

    assert asdict(OuterInput(nested=Some(None))) == {"nested": None}


def test_nested_input_with_maybe_field():
    @strawberry.input
    class InnerInput:
        required_field: str
        optional_field: Maybe[str | None]

    @strawberry.input
    class OuterInput:
        nested: InnerInput

    assert asdict(OuterInput(nested=InnerInput(required_field="foo"))) == {
        "nested": {"required_field": "foo"}
    }
    assert asdict(
        OuterInput(nested=InnerInput(required_field="foo", optional_field=Some("bar")))
    ) == {"nested": {"required_field": "foo", "optional_field": "bar"}}


def test_maybe_nested_input_with_unset_inner_field():
    @strawberry.input
    class InnerInput:
        required_field: str
        optional_field: str | None

    @strawberry.input
    class OuterInput:
        nested: Maybe[InnerInput]

    assert asdict(
        OuterInput(nested=Some(InnerInput(required_field="foo", optional_field=UNSET)))
    ) == {"nested": {"required_field": "foo"}}


def test_empty_list():
    @strawberry.input
    class Input:
        items: list[str]

    assert asdict(Input(items=[])) == {"items": []}


def test_empty_tuple():
    @strawberry.input
    class Input:
        items: tuple[str, ...]

    assert asdict(Input(items=())) == {"items": ()}


def test_list_with_none_items():
    @strawberry.input
    class Input:
        items: list[str | None]

    assert asdict(Input(items=["a", None, "b"])) == {"items": ["a", None, "b"]}


def test_unset_inside_list_is_filtered():
    # UNSET values nested inside a list should be stripped, the same way
    # UNSET top-level fields are omitted from the output dict.
    @strawberry.input
    class Input:
        items: list[str | None]

    result = asdict(Input(items=[UNSET, "foo", UNSET]))
    assert result == {"items": ["foo"]}


def test_list_of_nested_inputs():
    @strawberry.input
    class InnerInput:
        field: str

    @strawberry.input
    class OuterInput:
        items: list[InnerInput]

    assert asdict(
        OuterInput(items=[InnerInput(field="foo"), InnerInput(field="bar")])
    ) == {"items": [{"field": "foo"}, {"field": "bar"}]}


def test_tuple_of_nested_inputs():
    @strawberry.input
    class InnerInput:
        field: str

    @strawberry.input
    class OuterInput:
        items: tuple[InnerInput, ...]

    assert asdict(
        OuterInput(items=(InnerInput(field="foo"), InnerInput(field="bar")))
    ) == {"items": ({"field": "foo"}, {"field": "bar"})}


def test_list_of_optional_nested_inputs():
    @strawberry.input
    class InnerInput:
        field: str

    @strawberry.input
    class OuterInput:
        items: list[InnerInput | None]

    assert asdict(
        OuterInput(items=[InnerInput(field="foo"), None, InnerInput(field="bar")])
    ) == {"items": [{"field": "foo"}, None, {"field": "bar"}]}


def test_enum_values_in_list():
    @strawberry.enum
    class Status(Enum):
        ACTIVE = "active"
        INACTIVE = "inactive"

    @strawberry.type
    class Result:
        statuses: list[Status]

    assert asdict(Result(statuses=[Status.ACTIVE, Status.INACTIVE])) == {
        "statuses": [Status.ACTIVE, Status.INACTIVE]
    }


def test_maybe_list():
    @strawberry.input
    class InnerInput:
        field: str

    @strawberry.input
    class OuterInput:
        items: Maybe[list[InnerInput]]

    assert asdict(
        OuterInput(items=Some([InnerInput(field="foo"), InnerInput(field="bar")]))
    ) == {"items": [{"field": "foo"}, {"field": "bar"}]}


def test_maybe_list_with_none_items():
    @strawberry.input
    class Input:
        items: Maybe[list[str | None]]

    assert asdict(Input(items=Some([None, "foo", None]))) == {
        "items": [None, "foo", None]
    }


def test_optional_list():
    @strawberry.input
    class Input:
        items: list[str] | None = None

    assert asdict(Input()) == {"items": None}
    assert asdict(Input(items=["a", "b"])) == {"items": ["a", "b"]}



import dataclasses
from collections.abc import (
    AsyncGenerator,
    AsyncIterable,
    AsyncIterator,
    Generator,
    Iterable,
    Iterator,
)
from typing import (
    Annotated,
    Any,
    NamedTuple,
    TypeAlias,
    cast,
)
from typing_extensions import Self

import strawberry
from strawberry import relay
from strawberry.permission import BasePermission
from strawberry.relay.utils import to_base64
from strawberry.types import Info


@strawberry.type
class Fruit(relay.Node):
    id: relay.NodeID[int]
    name: str
    color: str

    @classmethod
    def resolve_nodes(
        cls,
        *,
        info: strawberry.Info,
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        if node_ids is not None:
            return [fruits[nid] if required else fruits.get(nid) for nid in node_ids]

        return fruits.values()

    @classmethod
    def is_type_of(cls, obj: Any, _info: strawberry.Info) -> bool:
        # This is here to support FruitConcrete, which is mimicing an integration
        # object which would return an object alike Fruit (e.g. the django integration)
        return isinstance(obj, (cls, FruitConcrete))


@dataclasses.dataclass
class FruitConcrete:
    id: int
    name: str
    color: str


@strawberry.type
class FruitAsync(relay.Node):
    id: relay.NodeID[int]
    name: str
    color: str

    @classmethod
    async def resolve_nodes(
        cls,
        *,
        info: Info | None = None,
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        if node_ids is not None:
            return [
                fruits_async[nid] if required else fruits_async.get(nid)
                for nid in node_ids
            ]

        return fruits_async.values()

    @classmethod
    async def resolve_id(cls, root: Self, *, info: strawberry.Info) -> str:
        return str(root.id)


@strawberry.type
class FruitCustomPaginationConnection(relay.Connection[Fruit]):
    @strawberry.field
    def something(self) -> str:
        return "foobar"

    @classmethod
    def resolve_connection(
        cls,
        nodes: Iterable[Fruit],
        *,
        info: Info | None = None,
        total_count: int | None = None,
        before: str | None = None,
        after: str | None = None,
        first: int | None = None,
        last: int | None = None,
        **kwargs: Any,
    ) -> Self:
        edges_mapping = {
            to_base64("fruit_name", n.name): relay.Edge(
                node=n,
                cursor=to_base64("fruit_name", n.name),
            )
            for n in sorted(nodes, key=lambda f: f.name)
        }
        edges = list(edges_mapping.values())
        first_edge = edges[0] if edges else None
        last_edge = edges[-1] if edges else None

        if after is not None:
            after_edge_idx = edges.index(edges_mapping[after])
            edges = [e for e in edges if edges.index(e) > after_edge_idx]

        if before is not None:
            before_edge_idx = edges.index(edges_mapping[before])
            edges = [e for e in edges if edges.index(e) < before_edge_idx]

        if first is not None:
            edges = edges[:first]

        if last is not None:
            edges = edges[-last:]

        return cls(
            edges=edges,
            page_info=relay.PageInfo(
                start_cursor=edges[0].cursor if edges else None,
                end_cursor=edges[-1].cursor if edges else None,
                has_previous_page=first_edge is not None
                and bool(edges)
                and edges[0] != first_edge,
                has_next_page=last_edge is not None
                and bool(edges)
                and edges[-1] != last_edge,
            ),
        )


fruits = {
    str(f.id): f
    for f in [
        Fruit(id=1, name="Banana", color="yellow"),
        Fruit(id=2, name="Apple", color="red"),
        Fruit(id=3, name="Pineapple", color="yellow"),
        Fruit(id=4, name="Grape", color="purple"),
        Fruit(id=5, name="Orange", color="orange"),
    ]
}
fruits_async = {
    k: FruitAsync(id=v.id, name=v.name, color=v.color) for k, v in fruits.items()
}


class FruitAlike(NamedTuple):
    id: int
    name: str
    color: str


@strawberry.type
class FruitAlikeConnection(relay.ListConnection[Fruit]):
    @classmethod
    def resolve_node(
        cls, node: FruitAlike, *, info: strawberry.Info, **kwargs: Any
    ) -> Fruit:
        return Fruit(
            id=node.id,
            name=node.name,
            color=node.color,
        )


def fruits_resolver() -> Iterable[Fruit]:
    return fruits.values()


async def fruits_async_resolver() -> Iterable[FruitAsync]:
    return fruits_async.values()


class DummyPermission(BasePermission):
    message = "Dummy message"

    async def has_permission(
        self, source: Any, info: strawberry.Info, **kwargs: Any
    ) -> bool:
        return True


FruitsListConnectionAlias: TypeAlias = relay.ListConnection[Fruit]


@strawberry.type
class Query:
    node: relay.Node = relay.node()
    node_with_async_permissions: relay.Node = relay.node(
        permission_classes=[DummyPermission]
    )
    nodes: list[relay.Node] = relay.node()
    node_optional: relay.Node | None = relay.node()
    nodes_optional: list[relay.Node | None] = relay.node()
    fruits: relay.ListConnection[Fruit] = relay.connection(resolver=fruits_resolver)
    fruits_lazy: relay.ListConnection[
        Annotated["Fruit", strawberry.lazy("tests.relay.schema")]
    ] = relay.connection(resolver=fruits_resolver)
    fruits_alias: FruitsListConnectionAlias = relay.connection(resolver=fruits_resolver)
    fruits_alias_lazy: Annotated[
        "FruitsListConnectionAlias",
        strawberry.lazy("tests.relay.schema"),
    ] = relay.connection(resolver=fruits_resolver)
    fruits_async: relay.ListConnection[FruitAsync] = relay.connection(
        resolver=fruits_async_resolver
    )
    fruits_custom_pagination: FruitCustomPaginationConnection = relay.connection(
        resolver=fruits_resolver
    )

    @relay.connection(relay.ListConnection[Fruit])
    def fruits_concrete_resolver(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> list[Fruit]:
        # This is mimicing integrations, like Django
        return [
            cast(
                "Fruit",
                FruitConcrete(
                    id=f.id,
                    name=f.name,
                    color=f.color,
                ),
            )
            for f in fruits.values()
            if name_endswith is None or f.name.endswith(name_endswith)
        ]

    @relay.connection(relay.ListConnection[Fruit])
    def fruits_custom_resolver(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> list[Fruit]:
        return [
            f
            for f in fruits.values()
            if name_endswith is None or f.name.endswith(name_endswith)
        ]

    @relay.connection(relay.ListConnection[Fruit])
    def fruits_custom_resolver_lazy(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> list[Annotated["Fruit", strawberry.lazy("tests.relay.schema")]]:
        return [
            f
            for f in fruits.values()
            if name_endswith is None or f.name.endswith(name_endswith)
        ]

    @relay.connection(relay.ListConnection[Fruit])
    def fruits_custom_resolver_iterator(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> Iterator[Fruit]:
        for f in fruits.values():
            if name_endswith is None or f.name.endswith(name_endswith):
                yield f

    @relay.connection(relay.ListConnection[Fruit])
    def fruits_custom_resolver_iterable(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> Iterator[Fruit]:
        for f in fruits.values():
            if name_endswith is None or f.name.endswith(name_endswith):
                yield f

    @relay.connection(relay.ListConnection[Fruit])
    def fruits_custom_resolver_generator(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> Generator[Fruit, None, None]:
        for f in fruits.values():
            if name_endswith is None or f.name.endswith(name_endswith):
                yield f

    @relay.connection(relay.ListConnection[Fruit])
    async def fruits_custom_resolver_async_iterable(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> AsyncIterable[Fruit]:
        for f in fruits.values():
            if name_endswith is None or f.name.endswith(name_endswith):
                yield f

    @relay.connection(relay.ListConnection[Fruit])
    async def fruits_custom_resolver_async_iterator(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> AsyncIterator[Fruit]:
        for f in fruits.values():
            if name_endswith is None or f.name.endswith(name_endswith):
                yield f

    @relay.connection(relay.ListConnection[Fruit])
    async def fruits_custom_resolver_async_generator(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> AsyncGenerator[Fruit, None]:
        for f in fruits.values():
            if name_endswith is None or f.name.endswith(name_endswith):
                yield f

    @relay.connection(FruitAlikeConnection)
    def fruit_alike_connection_custom_resolver(
        self,
        info: strawberry.Info,
        name_endswith: str | None = None,
    ) -> list[FruitAlike]:
        return [
            FruitAlike(f.id, f.name, f.color)
            for f in fruits.values()
            if name_endswith is None or f.name.endswith(name_endswith)
        ]

    @strawberry.relay.connection(strawberry.relay.ListConnection[Fruit])
    def some_fruits(self) -> list[Fruit]:
        return [Fruit(id=x, name="apple", color="green") for x in range(200)]


@strawberry.type
class CreateFruitPayload:
    fruit: Fruit


@strawberry.type
class Mutation:
    @strawberry.mutation
    def create_fruit(
        self,
        info: strawberry.Info,
        name: str,
        color: str,
    ) -> CreateFruitPayload: ...


schema = strawberry.Schema(query=Query, mutation=Mutation)



import enum
import textwrap
from typing import Annotated

import strawberry
from strawberry.printer import print_schema


@strawberry.enum
class LazyEnum(enum.Enum):
    BREAD = "BREAD"


def test_lazy_enum():
    @strawberry.type
    class Query:
        a: Annotated[
            "LazyEnum", strawberry.lazy("tests.schema.test_lazy_types.test_lazy_enums")
        ]

    expected = """
    enum LazyEnum {
      BREAD
    }

    type Query {
      a: LazyEnum!
    }
    """

    schema = strawberry.Schema(Query)

    assert print_schema(schema) == textwrap.dedent(expected).strip()



from .clients.base import HttpClient


async def test_mutation(http_client: HttpClient):
    response = await http_client.query(
        query="mutation { hello }",
        headers={
            "Content-Type": "application/json",
        },
    )
    data = response.json["data"]

    assert response.status_code == 200
    assert isinstance(data, dict)
    assert data["hello"] == "strawberry"



# type: ignore
from __future__ import annotations

import textwrap
from typing import Annotated

import strawberry
from strawberry.printer import print_schema
from strawberry.scalars import JSON
from strawberry.types.base import StrawberryList, StrawberryOptional
from tests.a import A
from tests.d import D


def test_forward_reference():
    global MyType

    @strawberry.type
    class Query:
        me: MyType = strawberry.field(name="myself")

    @strawberry.type
    class MyType:
        id: strawberry.ID
        scalar: JSON
        optional_scalar: JSON | None

    expected_representation = '''
    """
    The `JSON` scalar type represents JSON values as specified by [ECMA-404](https://ecma-international.org/wp-content/uploads/ECMA-404_2nd_edition_december_2017.pdf).
    """
    scalar JSON @specifiedBy(url: "https://ecma-international.org/wp-content/uploads/ECMA-404_2nd_edition_december_2017.pdf")

    type MyType {
      id: ID!
      scalar: JSON!
      optionalScalar: JSON
    }

    type Query {
      myself: MyType!
    }
    '''

    schema = strawberry.Schema(Query)

    assert print_schema(schema) == textwrap.dedent(expected_representation).strip()

    del MyType


def test_lazy_forward_reference():
    @strawberry.type
    class Query:
        @strawberry.field
        async def a(self) -> A:  # pragma: no cover
            return A(id=strawberry.ID("1"))

    expected_representation = """
    type A {
      id: ID!
      b: B!
      optionalB: B
      optionalB2: B
    }

    type B {
      id: ID!
      a: A!
      aList: [A!]!
      optionalA: A
      optionalA2: A
    }

    type Query {
      a: A!
    }
    """

    schema = strawberry.Schema(query=Query)
    assert print_schema(schema) == textwrap.dedent(expected_representation).strip()


def test_lazy_forward_reference_schema_with_a_list_only():
    @strawberry.type
    class Query:
        @strawberry.field
        async def d(self) -> D:  # pragma: no cover
            return D(id=strawberry.ID("1"))

    expected_representation = """
    type C {
      id: ID!
    }

    type D {
      id: ID!
      cList: [C!]!
    }

    type Query {
      d: D!
    }
    """

    schema = strawberry.Schema(query=Query)
    assert print_schema(schema) == textwrap.dedent(expected_representation).strip()


def test_with_resolver():
    global User

    @strawberry.type
    class User:
        name: str

    def get_users() -> list[User]:  # pragma: no cover
        return []

    @strawberry.type
    class Query:
        users: list[User] = strawberry.field(resolver=get_users)

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field] = definition.fields

    assert field.python_name == "users"
    assert isinstance(field.type, StrawberryList)
    assert field.type.of_type is User

    del User


def test_union_or_notation():
    global User

    @strawberry.type
    class User:
        name: str

    def get_users() -> list[User] | None:  # pragma: no cover
        return []

    @strawberry.type
    class Query:
        users: list[User] | None = strawberry.field(resolver=get_users)

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field] = definition.fields

    assert field.python_name == "users"
    assert isinstance(field.type, StrawberryOptional)
    assert isinstance(field.type.of_type, StrawberryList)
    assert field.type.of_type.of_type is User

    del User


def test_union_or_notation_generic_type_alias():
    global User

    @strawberry.type
    class User:
        name: str

    def get_users() -> list[User] | None:  # pragma: no cover
        return []

    @strawberry.type
    class Query:
        users: list[User] | None = strawberry.field(resolver=get_users)

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field] = definition.fields

    assert field.python_name == "users"
    assert isinstance(field.type, StrawberryOptional)
    assert isinstance(field.type.of_type, StrawberryList)
    assert field.type.of_type.of_type is User

    del User


def test_annotated():
    global User

    @strawberry.type
    class User:
        name: str

    def get_users() -> list[User]:  # pragma: no cover
        return []

    @strawberry.type
    class Query:
        users: Annotated[list[User], object()] = strawberry.field(resolver=get_users)

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field] = definition.fields

    assert field.python_name == "users"
    assert isinstance(field.type, StrawberryList)
    assert field.type.of_type is User

    del User


def test_annotated_or_notation():
    global User

    @strawberry.type
    class User:
        name: str

    def get_users() -> list[User] | None:  # pragma: no cover
        return []

    @strawberry.type
    class Query:
        users: Annotated[list[User] | None, object()] = strawberry.field(
            resolver=get_users
        )

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field] = definition.fields

    assert field.python_name == "users"
    assert isinstance(field.type, StrawberryOptional)
    assert isinstance(field.type.of_type, StrawberryList)
    assert field.type.of_type.of_type is User

    del User


def test_annotated_or_notation_generic_type_alias():
    global User

    @strawberry.type
    class User:
        name: str

    def get_users() -> list[User]:  # pragma: no cover
        return []

    @strawberry.type
    class Query:
        users: Annotated[list[User] | None, object()] = strawberry.field(
            resolver=get_users
        )

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field] = definition.fields

    assert field.python_name == "users"
    assert isinstance(field.type, StrawberryOptional)
    assert isinstance(field.type.of_type, StrawberryList)
    assert field.type.of_type.of_type is User

    del User



from __future__ import annotations

import contextlib
import json
from collections.abc import AsyncGenerator, Sequence
from datetime import timedelta
from io import BytesIO
from typing import Any, Literal

from fastapi import BackgroundTasks, Depends, FastAPI, Request, WebSocket
from fastapi.testclient import TestClient

from strawberry.fastapi import GraphQLRouter as BaseGraphQLRouter
from strawberry.http import GraphQLHTTPResponse
from strawberry.http.ides import GraphQL_IDE
from strawberry.schema import Schema
from strawberry.subscriptions import (
    GRAPHQL_TRANSPORT_WS_PROTOCOL,
    GRAPHQL_WS_PROTOCOL,
)
from strawberry.types import ExecutionResult
from tests.http.context import get_context
from tests.views.schema import Query
from tests.websockets.views import OnWSConnectMixin

from .asgi import AsgiWebSocketClient
from .base import (
    JSON,
    DebuggableGraphQLTransportWSHandler,
    DebuggableGraphQLWSHandler,
    HttpClient,
    Response,
    ResultOverrideFunction,
    WebSocketClient,
)


def custom_context_dependency() -> str:
    return "Hi!"


def fastapi_get_context(
    background_tasks: BackgroundTasks,
    request: Request = None,  # type: ignore
    ws: WebSocket = None,  # type: ignore
    custom_value: str = Depends(custom_context_dependency),
) -> dict[str, object]:
    return get_context(
        {
            "request": request or ws,
            "background_tasks": background_tasks,
        }
    )


def get_root_value(
    request: Request = None,  # type: ignore - FastAPI
    ws: WebSocket = None,  # type: ignore - FastAPI
) -> Query:
    return Query()


class GraphQLRouter(OnWSConnectMixin, BaseGraphQLRouter[dict[str, object], object]):
    result_override: ResultOverrideFunction = None
    graphql_transport_ws_handler_class = DebuggableGraphQLTransportWSHandler
    graphql_ws_handler_class = DebuggableGraphQLWSHandler

    async def process_result(
        self, request: Request, result: ExecutionResult
    ) -> GraphQLHTTPResponse:
        if self.result_override:
            return self.result_override(result)

        return await super().process_result(request, result)


class FastAPIHttpClient(HttpClient):
    def __init__(
        self,
        schema: Schema,
        graphql_ide: GraphQL_IDE | None = "graphiql",
        allow_queries_via_get: bool = True,
        keep_alive: bool = False,
        keep_alive_interval: float = 1,
        subscription_protocols: Sequence[str] = (
            GRAPHQL_TRANSPORT_WS_PROTOCOL,
            GRAPHQL_WS_PROTOCOL,
        ),
        connection_init_wait_timeout: timedelta = timedelta(minutes=1),
        result_override: ResultOverrideFunction = None,
        multipart_uploads_enabled: bool = False,
        max_subscriptions_per_connection: int | None = 100,
    ):
        self.app = FastAPI()

        graphql_app = GraphQLRouter(
            schema,
            graphql_ide=graphql_ide,
            allow_queries_via_get=allow_queries_via_get,
            keep_alive=keep_alive,
            keep_alive_interval=keep_alive_interval,
            subscription_protocols=subscription_protocols,
            connection_init_wait_timeout=connection_init_wait_timeout,
            multipart_uploads_enabled=multipart_uploads_enabled,
            max_subscriptions_per_connection=max_subscriptions_per_connection,
            context_getter=fastapi_get_context,
            root_value_getter=get_root_value,
        )
        graphql_app.result_override = result_override
        self.app.include_router(graphql_app, prefix="/graphql")

        self.client = TestClient(self.app)

    async def _handle_response(self, response: Any) -> Response:
        # TODO: here we should handle the stream
        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=response.headers,
        )

    async def _graphql_request(
        self,
        method: Literal["get", "post"],
        query: str | None = None,
        operation_name: str | None = None,
        variables: dict[str, object] | None = None,
        files: dict[str, BytesIO] | None = None,
        headers: dict[str, str] | None = None,
        extensions: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response:
        body = self._build_body(
            query=query,
            operation_name=operation_name,
            variables=variables,
            files=files,
            method=method,
            extensions=extensions,
        )

        if body:
            if method == "get":
                kwargs["params"] = body
            elif files:
                kwargs["data"] = body
            else:
                kwargs["content"] = json.dumps(body)

        if files:
            kwargs["files"] = files

        response = getattr(self.client, method)(
            "/graphql",
            headers=self._get_headers(method=method, headers=headers, files=files),
            **kwargs,
        )

        return await self._handle_response(response)

    async def request(
        self,
        url: str,
        method: Literal["head", "get", "post", "patch", "put", "delete"],
        headers: dict[str, str] | None = None,
    ) -> Response:
        response = getattr(self.client, method)(url, headers=headers)

        return await self._handle_response(response)

    async def get(
        self,
        url: str,
        headers: dict[str, str] | None = None,
    ) -> Response:
        return await self.request(url, "get", headers=headers)

    async def post(
        self,
        url: str,
        data: bytes | None = None,
        json: JSON | None = None,
        headers: dict[str, str] | None = None,
    ) -> Response:
        response = self.client.post(url, headers=headers, content=data, json=json)

        return await self._handle_response(response)

    @contextlib.asynccontextmanager
    async def ws_connect(
        self,
        url: str,
        *,
        protocols: list[str],
    ) -> AsyncGenerator[WebSocketClient, None]:
        with self.client.websocket_connect(url, protocols) as ws:
            yield AsgiWebSocketClient(ws)
